import shutil
import math
import csv
from annogesiclib.gff3 import Gff3Parser
from annogesiclib.helper import Helper


def read_file(seq_file, srna_table):
    seq = ""
    with open(seq_file) as fh:
        for line in fh:
            if not line.startswith(">"):
                line = line.strip()
                seq = seq + line
    tabs = []
    sh = open(srna_table, "r")
    for row in csv.reader(sh, delimiter='\t'):
        tabs.append({"info": row, "seq_id": row[0],
                     "start": int(row[2]), "end": int(row[3]),
                     "strand": row[4]})
    return seq, tabs

def get_table_entry(tabs, srna):
    for tab in tabs:
        if (srna.seq_id == tab["seq_id"]) and (
                srna.start == tab["start"]) and (
                srna.end == tab["end"]) and (
                srna.strand == tab["strand"]):
            return tab

def backward_t(seq, start, end, strand, mut_u):
    no_ut = 0
    ext = 0
    bord = end - start
    while 1:
        if strand == "+":
            nt = Helper().extract_gene(seq, end - ext, end - ext, strand)
        else:
            nt = Helper().extract_gene(seq, start + ext,
                                       start + ext, strand)
        if (nt == "U") or (nt == "T"):
            pass
        else:
            no_ut += 1
            if no_ut > mut_u:
                break
        ext += 1
        if ext >= bord:
            break
    return ext    

def forward_t(seq, start, end, strand, mut_u):
    no_ut = 0
    ext = 0
    bord = end - start
    while 1:
        if strand == "+":
            nt = Helper().extract_gene(seq, end + ext, end + ext, strand)
        else:
            nt = Helper().extract_gene(seq, start - ext,
                                       start - ext, strand)
        if (nt == "U") or (nt == "T"):
            pass
        else:
            no_ut += 1
            if no_ut > mut_u:
                break
        ext += 1
        if ext >= bord:
            break
    return ext

def iterate_seq(seq_u, args_srna):
    pos = 0
    first = True
    nts = {"ut": 0, "no_ut": 0}
    while 1:
        if (len(seq_u) - pos) < args_srna.num_u:
            break
        for nt in reversed(seq_u[:(len(seq_u) - pos)]):
            if first:
                first = False
                if (nt != "U") and (nt != "T"):
                    break
            if (nt == "U") or (nt == "T"):
                nts["ut"] += 1
            else:
                nts["no_ut"] += 1
                if nts["no_ut"] > args_srna.mut_u:
                    break
        if nts["ut"] < args_srna.num_u:
            nts = {"ut": 0, "no_ut": 0}
            first = True
        else:
            break
        pos += 1
    return pos


def search_t(seq, start, end, strand, ext_b, ext_f, args_srna):
    if strand == "+":
        seq_end = end + args_srna.len_u + ext_f + 1
        if seq_end > len(seq):
            seq_end = len(seq)
        seq_u = Helper().extract_gene(
                    seq, end - ext_b - 1, seq_end, strand)
    else:
        seq_start = start - args_srna.len_u - ext_f - 1
        if (seq_start) < 1:
            seq_start = 1
        seq_u = Helper().extract_gene(seq, seq_start,
                                      start + ext_b + 1, strand)
    pos = iterate_seq(seq_u, args_srna)
    if strand == "+":
        final_end = (seq_end - pos)
        if (final_end - end) <= 0:
            final_end = end
        elif (final_end - start) >= args_srna.max_len:
            diff = final_end - start - args_srna.max_len
            pos = iterate_seq(seq_u[:(final_end - diff)], args_srna)
            final_end = (final_end - diff - pos)
            if (final_end - end) <= 0:
                final_end = end
        final_start = start
    else:
        final_start = (seq_start + pos)
        if (start - final_start) <= 0:
            final_start = start
        elif (end - final_start) >= args_srna.max_len:
            diff = end - final_start - args_srna.max_len
            pos = iterate_seq(seq_u[(final_start + diff):], args_srna)
            final_start = (final_start + diff + pos)
            if (start - final_start) <= 0:
                final_start = start
        final_end = end
    return final_start, final_end

def check_term(srna, tab, seq, len_u, out, out_t):
    if "with_term" in srna.attributes.keys():
        feature = srna.attributes["with_term"].split(":")[0]
        info = srna.attributes["with_term"].split(":")[-1]
        start = int(info.split("-")[0])
        end = int(info.split("-")[-1].split("_")[0])
        strand = info.split("_")[-1]

def get_srna_poly_u(srna_file, seq_file, srna_table, args_srna):
    seq, tabs = read_file(seq_file, srna_table)
    out = open(srna_file + "_ext_polyu", "w")
    out_t = open(srna_table + "_ext_polyu", "w")
    gff_f = open(srna_file, "r")
    for entry in Gff3Parser().entries(gff_f):
        tab = get_table_entry(tabs, entry)
        ext_b = backward_t(seq, entry.start, entry.end,
                           entry.strand, args_srna.mut_u)
        ext_f = forward_t(seq, entry.start - args_srna.len_u,
                          entry.end + args_srna.len_u, entry.strand,
                          args_srna.mut_u)
        final_start, final_end = search_t(seq, entry.start, entry.end,
                                          entry.strand, ext_b, ext_f,
                                          args_srna)
        tab["info"][2] = str(final_start)
        tab["info"][3] = str(final_end)
        info_without_attributes = "\t".join([str(field) for field in [
                    entry.seq_id, entry.source, entry.feature, str(final_start),
                    str(final_end), entry.score, entry.strand, entry.phase]])
        out.write("\t".join([info_without_attributes,
                             entry.attribute_string]) + "\n")
        out_t.write("\t".join(tab["info"]) + "\n")
    gff_f.close()
    out.close()
    out_t.close()
    shutil.move(srna_file + "_ext_polyu", srna_file)
    shutil.move(srna_table + "_ext_polyu", srna_table)



import csv
import shutil
from annogesiclib.lib_reader import read_libs


def import_covers(row):
    cover_names = []
    covers = []
    for data in row.split("("):
        if ")" not in data:
            cover_names.append(data)
        else:
            covers.append(data.split(")")[0])
            if len(data.split(");")) == 2:
                cover_names.append(data.split(");")[-1])
    return cover_names, covers

def get_lib_name(libs):
    tracks = []
    double_tracks = []
    track_list = []
    for lib1 in libs:
        for lib2 in libs:
            if (lib1["cond"] == lib2["cond"]) and (
                    lib1["type"] == lib2["type"]) and (
                    lib1["rep"] == lib2["rep"]) and (
                    lib1["strand"] != lib2["strand"]):
                track = "/".join([lib1["name"], lib2["name"]])
                if track not in double_tracks:
                    double_tracks.append(track)
                    double_tracks.append("/".join([lib2["name"],
                                         lib1["name"]]))
                    tracks.append(track)
                    track_list.append([lib1["name"], lib2["name"]])
    return tracks, track_list

def reorganize_table(input_libs, wigs, cover_header, table_file):
    libs, texs = read_libs(input_libs, wigs)
    fh = open(table_file, "r")
    first = True
    headers = []
    tracks, track_list = get_lib_name(libs)
    out = open(table_file + "tmp", "w")
    for row in csv.reader(fh, delimiter='\t'):
        if first:
            detect = False
            header_num = 0
            for header in row:
                if header == cover_header:
                    index = header_num
                    detect = True
                header_num += 1
                if not detect:
                   headers.append(header)
                else:
                   detect = False
            first = False
            for track in tracks:
                headers.append("Avg_coverage:" + track)
            out.write("\t".join(headers) + "\n")
        else:
            if len(row) < (index + 1):
                cover_names = []
                covers = []
            else:
                cover_names, covers = import_covers(row[index])
            if len(row) == index + 1:
                row = row[:index]
            else:
                row = row[:index] + row[index + 1:]
            detects = ["Not_detect"] * len(tracks)
            for name, cover in zip(cover_names, covers):
                num_track = 0
                for track in track_list:
                    if name in track:
                        detects[num_track] = cover
                    num_track += 1
            out.write("\t".join(row + detects) + "\n")
    out.close()
    shutil.move(table_file + "tmp", table_file)



class WigParser(object):
    '''parser the wiggle file based on 
    strain, track, position and coverage'''

    def parser(self, wig_fh, strand):
        track = ""
        strain = ""
        for line in wig_fh.readlines():
            line = line.strip()
            if (len(line) != 0) and (not line.startswith("#")):
                datas = line.split(" ")
                if (datas[0] == "variableStep") or (datas[0] == "fixedStep"):
                    strain = datas[1].split("=")
                    strain = strain[1].strip()
                    pre_pos = 0
                    first = True
                if (datas[0] == "track"):
                    track = datas[2].split("=")
                    track = track[1].replace("\"", "")
                    pre_pos = 0
                    first = True
                if (datas[0] != "track") and (
                        datas[0] != "variableStep") and (
                        datas[0] != "fixedStep"):
                    if len(datas) != 2:
                        datas = line.split("\t")
                    if int(datas[0]) - 1 != pre_pos:
                        for pos in range(pre_pos + 1, int(datas[0])):
                            yield AssignValue(pos, 0, strand, strain, track)
                        pre_pos = int(datas[0])
                        first = True
                    if (int(datas[0]) - 1 == pre_pos) or (first):
                        pre_pos = int(datas[0])
                        first = False
                        yield AssignValue(datas[0], datas[1],
                                           strand, strain, track)


class AssignValue(object):

    def __init__(self, pos, coverage, strand, strain, track):
        self.pos = int(pos)
        if strand == "+":
            self.coverage = float(coverage)
        else:
            if float(coverage) < 0:
                self.coverage = -1 * float(coverage)
            else:
                self.coverage = float(coverage)
        self.strand = strand
        self.strain = strain
        self.track = track

    def __str__(self):
        return "{0} {1} {2} {3} {4}".format(
                self.pos, self.coverage, self.strand, self.strain, self.track)



import os
import sys
import shutil
from subprocess import call, DEVNULL
from annogesiclib.multiparser import Multiparser
from annogesiclib.converter import Converter
from annogesiclib.format_fixer import FormatFixer
from annogesiclib.helper import Helper


class RATT(object):
    '''annotation transfer'''

    def __init__(self, args_ratt):
        self.multiparser = Multiparser()
        self.converter = Converter()
        self.format_fixer = FormatFixer()
        self.helper = Helper()
        if args_ratt.ref_gbk:
            self.gbk = os.path.join(args_ratt.ref_gbk, "gbk_tmp")
            self.gbk_tmp = os.path.join(self.gbk, "tmp")
            self.embl = os.path.join(args_ratt.ref_gbk, "embls")
        if args_ratt.ref_embls:
            self.embl = args_ratt.ref_embls
        self.ratt_log = os.path.join(args_ratt.output_path, "ratt_log.txt")
        self.tmp_files = {"tar": os.path.join(args_ratt.tar_fastas, "tmp"),
                          "ref": os.path.join(args_ratt.ref_fastas, "tmp"),
                          "out_gff": os.path.join(args_ratt.gff_outfolder,
                                                  "tmp"),
                          "gff": os.path.join(args_ratt.gff_outfolder,
                                              "tmp.gff"),
                          "ptt": os.path.join(args_ratt.gff_outfolder,
                                              "tmp.ptt"),
                          "rnt": os.path.join(args_ratt.gff_outfolder,
                                              "tmp.rnt")}

    def _convert_to_pttrnt(self, gffs, files, log):
        for gff in files:
            if gff.endswith(".gff"):
                gff = os.path.join(gffs, gff)
                filename = gff.split("/")
                prefix = filename[-1][:-4]
                rnt = gff[:-3] + "rnt"
                ptt = gff[:-3] + "ptt"
                fasta = self.helper.get_correct_file(self.tmp_files["tar"],
                                                     ".fa", prefix, None, None)
                if fasta:
                    self.converter.convert_gff2rntptt(gff, fasta, ptt, rnt,
                                                      None, None)
                    log.write("\t" + ptt + " is generated.\n")
                    log.write("\t" + rnt + " is generated.\n")

    def _remove_files(self, args_ratt, out_gbk, log):
        self.helper.remove_all_content(args_ratt.gff_outfolder, ".gff", "file")
        self.helper.remove_all_content(args_ratt.gff_outfolder, ".ptt", "file")
        self.helper.remove_all_content(args_ratt.gff_outfolder, ".rnt", "file")
        log.write("Moving the final output files to {0}.\n".format(args_ratt.gff_outfolder))
        self.helper.move_all_content(self.tmp_files["out_gff"],
                                     args_ratt.gff_outfolder, None)
        log.write("Remove the temperary files.\n")
        shutil.rmtree(self.tmp_files["out_gff"])
        shutil.rmtree(self.tmp_files["tar"])
        shutil.rmtree(self.tmp_files["ref"])
        self.helper.remove_tmp_dir(args_ratt.tar_fastas)
        self.helper.remove_tmp_dir(args_ratt.ref_fastas)
        self.helper.remove_tmp_dir(args_ratt.ref_embls)
        self.helper.remove_tmp_dir(args_ratt.ref_gbk)

    def _convert_to_gff(self, ratt_result, args_ratt, files, log):
        name = ratt_result.split(".")
        filename = ".".join(name[1:-2]) + ".gff"
        output_file = os.path.join(args_ratt.output_path, filename)
        self.converter.convert_embl2gff(
             os.path.join(args_ratt.output_path, ratt_result), output_file)
        self.format_fixer.fix_ratt(output_file, ".".join(name[1:-2]),
                                   "tmp_gff")
        shutil.move("tmp_gff", output_file)
        shutil.copy(output_file, os.path.join(args_ratt.gff_outfolder,
                                              filename))
        log.write("\t" + os.path.join(args_ratt.gff_outfolder, filename) + 
                  " is generated.\n")
        files.append(filename)

    def _parser_embl_gbk(self, files):
        self.helper.check_make_folder(self.gbk)
        for file_ in files:
            close = False
            with open(file_, "r") as f_h:
                for line in f_h:
                    if (line.startswith("LOCUS")):
                        out = open(self.gbk_tmp, "w")
                        datas = line.split(" ")
                        for data in datas:
                            if (len(data) != 0) and (data != "LOCUS"):
                                filename = ".".join([data.strip(), "gbk"])
                                break
                    elif (line.startswith("VERSION")):
                        datas = line.split(" ")
                        for data in datas:
                            if (len(data) != 0) and (data != "VERSION"):
                                new_filename = ".".join([data.strip(), "gbk"])
                                break
                        if new_filename.find(filename):
                            filename = new_filename
                    if out:
                        out.write(line)
                    if line.startswith("//"):
                        out.close()
                        close = True
                        shutil.move(self.gbk_tmp,
                                    os.path.join(self.gbk, filename))
            if not close:
                out.close()
        return self.gbk

    def _convert_embl(self, ref_embls, log):
        '''convert gbk to embl'''
        detect_gbk = False
        gbks = []
        out_gbk = None
        for embl in os.listdir(ref_embls):
            if (embl.endswith(".gbk")) or (
                    embl.endswith(".gbff")) or (
                    embl.endswith(".gb")):
                detect_gbk = True
                gbks.append(os.path.join(ref_embls, embl))
        if not detect_gbk:
            log.write("--related_gbk_files is assigned, but not gbk files are detected.\n"
                      "The gbk file names need to be ended at .gbk, .gb, or .gbff. \n")
            print("Error: Please assign proper Genebank files!")
            sys.exit()
        elif detect_gbk:
            out_gbk = self._parser_embl_gbk(gbks)
            log.write("Running converter.py to convert gbk file to embl format.\n")
            self.converter.convert_gbk2embl(out_gbk)
            self.helper.check_make_folder(self.embl)
            self.helper.move_all_content(out_gbk, self.embl, [".embl"])
            log.write("\t" + self.embl + " is generated and the embl files are stored in it.\n")
        return out_gbk

    def _run_ratt(self, args_ratt, tar, ref, out, log):
        if (not os.path.exists(self.embl)) or (
                not os.path.exists(os.path.join(
                    self.tmp_files["tar"], tar + ".fa"))) or (
                not os.path.exists(os.path.join(
                    self.tmp_files["ref"], ref + ".fa"))):
            print("Error: Please check --compare_pair, the strain names "
                  "should be the same as the strain names in fasta, "
                  "genbank or embl files!")
            log.write("The strain names in --compare_pair should be the same "
                      "as the strain names in fasta, genbank, or embl files.\n")
            sys.exit()
        log.write("Make sure your RATT version is at least 1.64.\n")
        log.write("If the RATT can not run properly, please check the "
                  "RATT_HOME and PAGIT_HOME is assigned correctly.\n")
        temp_embl_folder = os.path.join(self.embl, ref)
        os.mkdir(temp_embl_folder)
        shutil.copy(os.path.join(self.embl, ref + ".embl"), os.path.join(self.embl, ref))
        log.write(" ".join([args_ratt.ratt_path, self.embl,
              os.path.join(self.tmp_files["tar"], tar + ".fa"),
              args_ratt.element, args_ratt.transfer_type,
              os.path.join(self.tmp_files["ref"], ref + ".fa")]) + "\n")
        call([args_ratt.ratt_path, temp_embl_folder,
              os.path.join(self.tmp_files["tar"], tar + ".fa"),
              args_ratt.element, args_ratt.transfer_type,
              os.path.join(self.tmp_files["ref"], ref + ".fa")],
              stdout=out, stderr=DEVNULL)
        shutil.rmtree(temp_embl_folder)
#        call([args_ratt.ratt_path, self.embl,
#              os.path.join(self.tmp_files["tar"], tar + ".fa"),
#              args_ratt.element, args_ratt.transfer_type,
#              os.path.join(self.tmp_files["ref"], ref + ".fa")],
#             stdout=out, stderr=DEVNULL)
        log.write("Done!\n")

    def _format_and_run(self, args_ratt, log):
        print("Running RATT")
        for pair in args_ratt.pairs:
            ref = pair.split(":")[0]
            tar = pair.split(":")[1]
            out = open(self.ratt_log, "w+")
            self._run_ratt(args_ratt, tar, ref, out, log)
            log.write("The following files are generatd:\n")
            for filename in os.listdir():
                if ("final" in filename):
                    log.write("\t" + filename + "\n")
                    shutil.move(filename, os.path.join(args_ratt.output_path,
                                                       filename))
                elif (args_ratt.element in filename) or (
                      "query" in filename) or (
                      "Reference" in filename) or (
                      "Query" in filename) or (
                      "Sequences" in filename):
                    log.write("\t" + filename + "\n")
                    if os.path.isfile(filename):
                        os.remove(filename)
                    if os.path.isdir(filename):
                        shutil.rmtree(filename)
        out.close()

    def annotation_transfer(self, args_ratt, log):
        self.multiparser.parser_fasta(args_ratt.tar_fastas)
        self.multiparser.parser_fasta(args_ratt.ref_fastas)
        out_gbk = None
        if args_ratt.ref_embls is None:
            out_gbk = self._convert_embl(args_ratt.ref_gbk, log)
        self._format_and_run(args_ratt, log)
        files = []
        for data in os.listdir(args_ratt.output_path):
            if "final.embl" in data:
                log.write("Running converter.py to convert embl "
                          "files in {0} to gff, ptt, and rnt format.\n".format(data))
                self._convert_to_gff(data, args_ratt, files, log)
                self._convert_to_pttrnt(args_ratt.gff_outfolder, files, log)
        self.helper.check_make_folder(self.tmp_files["out_gff"])
        log.write("Merging the output of {0}.\n".format(data))
        for folder in os.listdir(args_ratt.tar_fastas):
            files = []
            if "_folder" in folder:
                datas = folder.split("_folder")
                prefix = ".".join(datas[0].split(".")[:-1])
                for file_ in os.listdir(os.path.join(args_ratt.tar_fastas,
                                                     folder)):
                    files.append(file_[:-3])
                for gff in os.listdir(args_ratt.gff_outfolder):
                    for file_ in files:
                        if (".gff" in gff) and (file_ == gff[:-4]):
                            self.helper.merge_file(os.path.join(
                                 args_ratt.gff_outfolder, gff),
                                 self.tmp_files["gff"])
                        if (".ptt" in gff) and (file_ == gff[:-4]):
                            self.helper.merge_file(os.path.join(
                                 args_ratt.gff_outfolder, gff),
                                 self.tmp_files["ptt"])
                        if (".rnt" in gff) and (file_ == gff[:-4]):
                            self.helper.merge_file(os.path.join(
                                 args_ratt.gff_outfolder, gff),
                                 self.tmp_files["rnt"])
                if os.path.exists(self.tmp_files["gff"]):
                    shutil.move(self.tmp_files["gff"], os.path.join(
                                self.tmp_files["out_gff"], prefix + ".gff"))
                    shutil.move(self.tmp_files["ptt"], os.path.join(
                                self.tmp_files["out_gff"], prefix + ".ptt"))
                    shutil.move(self.tmp_files["rnt"], os.path.join(
                                self.tmp_files["out_gff"], prefix + ".rnt"))
                else:
                    print("Error: Please check your fasta or "
                          "annotation files, they should only contain "
                          "the query genome. And make sure your RATT can "
                          "work properly (check $ANNOgesic/output/"
                          "annotation_transfer/ratt_log.txt).")
                    log.write("Please check your fasta or "
                              "annotation files, they should only contain "
                              "the query genome. And make sure your RATT can "
                              "work properly (check $ANNOgesic/output/"
                              "annotation_transfer/ratt_log.txt).\n")
        self._remove_files(args_ratt, out_gbk, log)



import csv
import shutil
from annogesiclib.helper import Helper
from annogesiclib.gff3 import Gff3Parser


def import_cds(gff):
    if "Name" in gff.attributes.keys():
        return gff.attributes["Name"]
    elif "ID" in gff.attributes.keys():
        return gff.attributes["ID"]
    else:
        name = "".join([gff.feature, ":", str(gff.start), "-", str(gff.end),
                        "_", gff.strand])
        return name


def check_overlap(table_file, gff_file):
    out = open(table_file + "tmp", "w")
    gffs = []
    gff_f = open(gff_file, "r")
    for entry in Gff3Parser().entries(gff_f):
        if Helper().feature_without_notgene(entry):
            gffs.append(entry)
    fh = open(table_file, "r")
    out.write("\t".join([
        "Rank", "Genome", "Name", "Start", "End", "Strand",
        "Start_with_TSS/Cleavage_site", "End_with_cleavage", "Candidates",
        "Lib_type", "Best_avg_coverage", "Track/Coverage",
        "Normalized_secondary_energy_change(by_length)", "sRNA_types",
        "Conflict_sORF", "nr_hit_number", "sRNA_hit_number",
        "nr_hit_top3|ID|e-value|score", "sRNA_hit|e-value|score", "Overlap_CDS_forward",
        "Overlap_nts_forward", "Overlap_CDS_reverse",
        "Overlap_nts_reverse","End_with_terminator",
        "Associated_promoter", "sRNA_length"]) + "\n")
    for row in csv.reader(fh, delimiter='\t'):
        if row[3] != "Start":
            overlaps = {"forward": [], "reverse": [],
                        "CDS_f": [], "CDS_r": []}
            start = int(row[3])
            end = int(row[4])
            for gff in gffs:
                if ((gff.end < end) and (
                         gff.end > start) and (
                         gff.start <= start)) or (
                        (gff.start > start) and (
                         gff.start < end) and (
                         gff.end >= end)) or (
                        (gff.end >= end) and (
                         gff.start <= start)) or (
                        (gff.end <= end) and (
                         gff.start >= start)):
                    overlap = min(gff.end, end) - max(gff.start, start) + 1
                    percent = "{0:.0f}%".format((float(overlap) / float(end - start + 1)) * 100)
                    if gff.strand == "+":
                        overlaps["forward"].append(str(overlap) + "(" + str(percent) + ")")
                        overlaps["CDS_f"].append(import_cds(gff))
                    else:
                        overlaps["reverse"].append(str(overlap) + "(" + str(percent) + ")")
                        overlaps["CDS_r"].append(import_cds(gff))
            if len(overlaps["forward"]) == 0:
                overlaps["forward"] = ["NA"]
                overlaps["CDS_f"] = ["NA"]
            if len(overlaps["reverse"]) == 0:
                overlaps["reverse"] = ["NA"]
                overlaps["CDS_r"] = ["NA"]
            out.write("\t".join(row[0:19] + [";".join(overlaps["CDS_f"]), ";".join(overlaps["forward"]),
                                             ";".join(overlaps["CDS_r"]), ";".join(overlaps["reverse"])] +
                                row[21:]) + "\n")
    shutil.move(table_file + "tmp", table_file)



import os
import sys
from annogesiclib.projectcreator import ProjectCreator
from annogesiclib.paths import Paths
from annogesiclib.get_input import get_file
from annogesiclib.converter import Converter
from annogesiclib.get_target_fasta import TargetFasta
from annogesiclib.ratt import RATT
from annogesiclib.tsspredator import TSSpredator
from annogesiclib.optimize import optimize_tss
from annogesiclib.color_png import ColorPNG
from annogesiclib.terminator import Terminator
from annogesiclib.transcript import TranscriptDetection
from annogesiclib.utr import UTRDetection
from annogesiclib.srna import sRNADetection
from annogesiclib.sorf import sORFDetection
from annogesiclib.meme import MEME
from annogesiclib.operon import OperonDetection
from annogesiclib.circrna import CircRNADetection
from annogesiclib.goterm import GoTermFinding
from annogesiclib.srna_target import sRNATargetPrediction
from annogesiclib.snp import SNPCalling
from annogesiclib.ppi import PPINetwork
from annogesiclib.sublocal import SubLocal
from annogesiclib.ribos import Ribos
from annogesiclib.crispr import Crispr
from annogesiclib.merge_feature import run_merge
from annogesiclib.overlap import deal_overlap
from annogesiclib.screen import Screen
from annogesiclib.args_container import ArgsContainer
from annogesiclib.helper import Helper
project_creator = ProjectCreator()


class Controller(object):

    """Manage the actions of the subcommands.

    The Controller take care of providing the argumentes like path
    names and the parallel processing of tasks.

    """
    def __init__(self, args):
        """Create an instance."""
        self._args = args
        if (len(args.__dict__) > 3):
            if not os.path.exists(args.project_path):
                print("Error: --project_path does not exists!")
                sys.exit()
        self._paths = Paths(args.project_path)
        self.args_container = ArgsContainer()
        self.helper = Helper()

    def check_folder(self, folders, flags, log):
        '''Check the emtpy or wrong assigned folder'''
        for folder, flag in zip(folders, flags):
            if folder is None:
                log.write("{0} of {1} is not found. Please check it!".format(
                          folder, flag))
                print("Error: {0} of {1} is not found. Please check it!".format(
                      folder, flag))
                sys.exit()
            else:
                if os.path.exists(folder):
                    if len(os.listdir(folder)) == 0:
                        log.write("{0} is a empty folder!".format(flag))
                        print("Error: {0} is a empty folder!".format(flag))
                        sys.exit()
                else:
                    log.write("{0} of {1} is not found. Please check it!".format(
                               folder, flag))
                    print("Error: {0} of {1} is not found. Please check it!".format(
                          folder, flag))
                    sys.exit()

    def check_multi_files(self, input_files, flags, log):
        if input_files is not None:
            for files, flag in zip(input_files, flags):
                if files is not None:
                    for file_ in files:
                        if not os.path.exists(file_):
                            print("Error: {0} in {1} does "
                                  "not exist!".format(file_, flag))
                            log.write(file_ + " does not exists\n")
                            sys.exit()
                        else:
                            log.write(file_ + " exists\n")

    def check_parameter(self, paras, names, log):
        '''Check the parameter is assigned correct or not'''
        for i in range(len(paras)):
            if paras[i] is None:
                print("Error: {0} can not be None. "
                      "Please check it!".format(names[i]))
                log.write(file_ + " need to be assigned.\n")
                sys.exit()

    def check_execute_file(self, exe, log):
        detect = False
        exe_folder = ""
        log.write("Checking " + exe + "\n")
        if os.path.exists(exe):
            detect = True
            full_exe = os.path.realpath(exe)
            log.write(full_exe + " is found.\n")
        else:
            exes = []
            for folder in os.environ["PATH"].split(":"):
                if os.path.isfile(os.path.join(folder, exe)):
                    exe_folder = folder
                    detect = True
                    full_exe = exe
                    if os.path.join(folder, exe) not in exes:
                        exes.append(os.path.join(folder, exe))
                        log.write(os.path.join(folder, exe) + " is found.\n")
        if not detect:
            if os.path.exists(os.path.realpath(exe)):
                full_exe = os.path.realpath(exe)
                log.write(full_exe + " is found.\n")
            else:
                print("Error: {0} can't be found!".format(exe))
                print("Please assign the correct path!")
                sys.exit()
        if (os.path.isfile(full_exe)) or (
                os.path.isfile(os.path.join(exe_folder, exe))):
            log.write("The execute path is " + os.popen(
                "which " + exe).read())
            return full_exe
        else:
            log.write(full_exe + " is not found.\n")
            print("Error: {0} is not a file!".format(exe))
            sys.exit()

    def check_file(self, files, names, require, log):
        '''Check the path of file'''
        for i in range(len(files)):
            if require:
                if files[i] is None:
                    print("Error: {0} can not be None. "
                          "Please check it!".format(names[i]))
                    log.write(names[i] + " is None.\n")
                    sys.exit()
                else:
                    if not os.path.isfile(files[i]):
                        print("Error: {0} is not found. "
                              "Please check it!".format(files[i]))
                        log.write(files[i] + " does not exist.\n")
                        sys.exit()
            else:
                if files[i] is not None:
                    if not os.path.isfile(files[i]):
                        print("Error: {0} is not found. "
                              "Please check it!".format(files[i]))
                        log.write(files[i] + " does not exist.\n")
                        sys.exit()

    def create_project(self, version):
        """Create a new project."""
        project_creator.create_root_folder(self._args.project_path)
        project_creator.create_subfolders(self._paths.required_folders("root"))
        project_creator.create_version_file(
            self._paths.version_path, version)
        sys.stdout.write("Created folder \"%s\" and required subfolders.\n" % (
            self._args.project_path))

    def get_input(self):
        """Download required files from website."""
        print("Running get input files")
        log = open(os.path.join(self._paths.reference_input_folder, "log.txt"), "w")
        if self._args.ftp_path is None:
            print("Error: Please assign the path for downloading the data!")
            sys.exit()
            annotation_folder = self._paths.ref_annotation_folder
            fasta_folder = self._paths.ref_fasta_folder
        self.helper.check_make_folder(self._paths.ref_annotation_folder)
        self.helper.check_make_folder(self._paths.ref_fasta_folder)
        if self._args.ref_gff is True:
            log.write("Get gff files\n")
            get_file(self._args.ftp_path, self._paths.ref_annotation_folder,
                     "gff", log)
            get_file(self._args.ftp_path, self._paths.ref_annotation_folder,
                     "_genomic.gff.gz", log)
        if self._args.ref_fasta is True:
            log.write("Get fasta files\n")
            get_file(self._args.ftp_path, self._paths.ref_fasta_folder,
                     "fna", log)
            get_file(self._args.ftp_path, self._paths.ref_fasta_folder,
                     "_genomic.fna.gz", log)
        if self._args.ref_gbk is True:
            log.write("Get gbk files\n")
            get_file(self._args.ftp_path, self._paths.ref_annotation_folder,
                     "gbk", log)
            get_file(self._args.ftp_path, self._paths.ref_annotation_folder,
                     "gbff", log)
            get_file(self._args.ftp_path, self._paths.ref_annotation_folder,
                     "_genomic.gbff.gz", log)
        if self._args.ref_ptt is True:
            log.write("Get ptt files\n")
            get_file(self._args.ftp_path, self._paths.ref_annotation_folder,
                     "ptt", log)
        if self._args.ref_rnt is True:
            log.write("Get rnt files\n")
            get_file(self._args.ftp_path, self._paths.ref_annotation_folder,
                     "rnt", log)
        if self._args.convert_embl is True:
            annotation_files = os.listdir(self._paths.ref_annotation_folder)
            if len(annotation_files) == 0:
                sys.stdout.write("No gff files!!\n")
            else:
                log.write("Running converter.py for converting gbk file "
                          "to embl formet\n")
                Converter().convert_gbk2embl(self._paths.ref_annotation_folder)
        log.close()

    def get_target_fasta(self):
        """Get target fasta"""
        print("Running update genome fasta")
        project_creator.create_subfolders(
            self._paths.required_folders("get_target_fasta"))
        log = open(os.path.join(self._paths.target_folder, "log.txt"), "w")
        self.check_multi_files([self._args.related_fasta_files],
                               ["--related_fasta_files"], log)
        self.check_file([self._args.mutation_table], ["--mutation_table"], True, log)
        target = TargetFasta(self._paths.tar_fasta_folder,
                             self._args.related_fasta_files)
        target.get_target_fasta(
                self._args.mutation_table, self._paths.tar_fasta_folder,
                self._args.related_fasta_files, self._args.updated_seq_name,
                self._paths.target_base_folder, log)
        log.close()

    def ratt(self):
        """Run RATT to transfer annotation file from reference to target."""
        print("Running annotation transfer")
        project_creator.create_subfolders(
            self._paths.required_folders("get_target_fasta"))
        project_creator.create_subfolders(
            self._paths.required_folders("annotation_transfer"))
        log = open(os.path.join(self._paths.ratt_folder, "log.txt"), "w")
        if (self._args.transfer_type != "Strain") and (
                self._args.transfer_type != "Assembly") and (
                self._args.transfer_type != "Species") and (
                self._args.transfer_type != "Assembly.Repetitive") and (
                self._args.transfer_type != "Strain.Repetitive") and (
                self._args.transfer_type != "Species.Repetitive") and (
                self._args.transfer_type != "Multiple") and (
                self._args.transfer_type != "Free"):
            log.write("Incorrect --transfer_type. Please assign 'Assembly', 'Species', "
                      "'Assembly.Repetitive', 'Strain.Repetitive', 'Species.Repetitive', "
                      "'Multiple' or 'Free'\n")
            print("Error: please assign correct --transfer_type!")
            sys.exit()
        if (self._args.related_embl_files is None) and (
                self._args.related_gbk_files is None):
            print("Error: please assign proper embl or genbank files")
            log.write("--related_gbk_files and --related_embl_files can not be both None.\n")
            sys.exit()
        elif (self._args.related_embl_files is not None) and (
                self._args.related_gbk_files is not None):
            log.write("Please choose --related_gbk_files as input or "
                      "--related_embl_files as input. Do not assign both.\n")
            print("Error: please choose embl as input or genbank as input")
            sys.exit()
        self._args.ratt_path = self.check_execute_file(self._args.ratt_path, log)
        self.check_multi_files(
                [self._args.target_fasta_files, self._args.related_fasta_files],
                ["--target_fasta_files", "--closed_fasta_files"], log)
        self.check_parameter([self._args.element, self._args.compare_pair],
                             ["--element", "--compare_pair"], log)
        args_ratt = self.args_container.container_ratt(
            self._args.ratt_path, self._args.element, self._args.transfer_type,
            self._args.related_embl_files, self._args.related_gbk_files,
            self._args.target_fasta_files, self._args.related_fasta_files,
            self._paths.ratt_folder,
            self._paths.tar_annotation_folder, self._args.compare_pair)
        ratt = RATT(args_ratt)
        ratt.annotation_transfer(args_ratt, log)
        log.close()


    def tsspredator(self):
        """Run TSSpredator for predicting TSS candidates."""
        if self._args.program.lower() == "tss":
            print("Running TSS prediction")
            project_creator.create_subfolders(
                self._paths.required_folders("TSS"))
            out_folder = self._paths.tsspredator_folder
            log = open(os.path.join(self._paths.tsspredator_folder,
                                    "log.txt"), "w")
            log.write("Running TSS prediction.\n")
        elif self._args.program.lower() == "ps":
            print("Running processing site prediction")
            out_folder = self._paths.processing_site_folder
            project_creator.create_subfolders(
                self._paths.required_folders("processing"))
            log = open(os.path.join(self._paths.processing_site_folder,
                                    "log.txt"), "w")
            log.write("Running PS prediction.\n")
        else:
            print("Error: No such program!")
            sys.exit()
        self.check_multi_files(
                [self._args.fasta_files, self._args.annotation_files,
                 self._args.compare_overlap_gff, self._args.manual_files,
                 self._args.compare_transcript_files],
                ["--fasta_files", "--annotation_files", "--compare_overlap_gff",
                 "--manual_files","--compare_transcript_files"], log)
        self.check_parameter([self._args.tex_notex_libs, self._args.condition_names],
                             ["--tex_notex_libs", "--condition_names"], log)
        self._args.tsspredator_path = self.check_execute_file(
                self._args.tsspredator_path, log)
        args_tss = self.args_container.container_tsspredator(
            self._args.tsspredator_path, self._args.program,
            self._args.fasta_files, self._args.annotation_files,
            self._args.tex_notex_libs, self._args.condition_names,
            self._args.output_id,
            self._args.auto_load_optimized_parameters, self._args.genome_order,
            self._args.height, self._args.height_reduction, self._args.factor,
            self._args.factor_reduction, self._args.base_height,
            self._args.enrichment_factor, self._args.processing_factor,
            self._args.replicate_tex, out_folder, self._args.validate_gene,
            self._args.manual_files, self._args.curated_sequence_length,
            self._args.compare_transcript_files, self._args.tolerance,
            self._args.utr_length, self._args.cluster,
            self._args.re_check_orphan,
            self._args.remove_overlap_feature, self._args.compare_overlap_gff,
            self._args.remove_low_expression)
        tsspredator = TSSpredator(args_tss)
        tsspredator.run_tsspredator(args_tss, log)
        log.close()

    def optimize(self):
        """opimize TSSpredator"""
        if self._args.program.lower() == "tss":
            print("Running optimization of TSS prediction")
            project_creator.create_subfolders(
                self._paths.required_folders("TSS"))
            out_folder = self._paths.tsspredator_folder
            if "optimized_TSSpredator" not in os.listdir(out_folder):
                os.mkdir(os.path.join(out_folder, "optimized_TSSpredator"))
            log = open(os.path.join(out_folder, "optimized_TSSpredator",
                                    "log.txt"), "w")
            log.write("Running optimization of TSS prediction\n")
        elif self._args.program.lower() == "ps":
            out_folder = self._paths.processing_site_folder
            project_creator.create_subfolders(
                self._paths.required_folders("processing"))
            if "optimized_TSSpredator" not in os.listdir(out_folder):
                os.mkdir(os.path.join(out_folder, "optimized_TSSpredator"))
            log = open(os.path.join(out_folder, "optimized_TSSpredator",
                                    "log.txt"), "w")
            log.write("Running optimization of PS prediction\n")
            print("Running optimization of processing site prediction")
        else:
            print("Error: No such program!")
            sys.exit()
        self.check_multi_files(
                [self._args.fasta_files, self._args.annotation_files,
                 self._args.manual_files],
                ["--fasta_files", "--annotation_files", "--manual_files"], log)
        self._args.tsspredator_path = self.check_execute_file(
                self._args.tsspredator_path, log)
        self.check_parameter([self._args.tex_notex_libs,
                              self._args.condition_names],
                             ["--tex_notex_lib",
                              "--condition_names"], log)
        args_ops = self.args_container.container_optimize(
            self._args.tsspredator_path, self._args.fasta_files,
            self._args.annotation_files,
            self._args.manual_files, out_folder, self._args.max_height,
            self._args.max_height_reduction, self._args.max_factor,
            self._args.max_factor_reduction, self._args.max_base_height,
            self._args.max_enrichment_factor, self._args.max_processing_factor,
            self._args.utr_length, self._args.tex_notex_libs,
            self._args.condition_names, self._args.output_id,
            self._args.cluster, self._args.curated_sequence_length,
            self._args.parallels, self._args.program,
            self._args.replicate_tex, self._args.steps)
        optimize_tss(args_ops, log)
        log.close()

    def color(self):
        """color the screenshots"""
        print("Running png files coloring")
        if not os.path.exists(os.path.join(self._args.screenshot_folder,
                                           "screenshots")):
            print("The folder -- screenshots needs to be found in "
                  "{0}.".format(self._args.screenshot_folder))
            sys.exit()
        log = open(os.path.join(self._args.screenshot_folder, "screenshots",
                                "color_log.txt"), "w")
        self.check_parameter([self._args.track_number], ["--track_numer"], log)
        self.check_folder([self._args.screenshot_folder], ["--screenshot_folder"], log)
        self._args.imagemagick_covert_path = self.check_execute_file(
                self._args.imagemagick_covert_path, log)
        color = ColorPNG()
        color.generate_color_png(
                self._args.track_number, self._args.screenshot_folder,
                self._args.imagemagick_covert_path, log)
        log.close()

    def terminator(self):
        """Run TransTermHP and Gene converaged for detecting terminators"""
        print("Running terminator prediction")
        project_creator.create_subfolders(
            self._paths.required_folders("terminator"))
        log = open(os.path.join(self._paths.transterm_folder, "log.txt"), "w")
        if self._args.transterm_path is None:
            print("Please assign the path of transterm in TransTermHP.")
        self.check_multi_files(
                [self._args.fasta_files, self._args.annotation_files,
                 self._args.transcript_files, self._args.srna_files],
                ["--fasta_files", "--annotation_files",
                 "--transcript_files", "--srna_files"], log)
        for prop in ("transterm_path", "expterm_path", "rnafold_path"):
            setattr(self._args, prop,
                    self.check_execute_file(getattr(self._args, prop), log))
        args_term = self.args_container.container_terminator(
            self._args.transterm_path, self._args.expterm_path,
            self._args.rnafold_path,
            self._paths.transterm_folder, self._args.fasta_files,
            self._args.annotation_files, self._args.transcript_files,
            self._args.srna_files, self._args.decrease,
            self._args.highest_coverage, self._args.tolerance_detect_coverage,
            self._args.tolerance_within_transcript,
            self._args.tolerance_downstream_transcript,
            self._args.tolerance_within_gene,
            self._args.tolerance_downstream_gene, self._paths.transtermhp_folder,
            self._args.tex_notex_libs, self._args.frag_libs,
            self._args.tex_notex, self._args.replicate_tex,
            self._args.replicate_frag,
            self._args.min_loop_length, self._args.max_loop_length,
            self._args.min_stem_length, self._args.max_stem_length,
            self._args.min_u_tail, self._args.miss_rate,
            self._args.mutation_u_tail, self._args.keep_multi_term,
            self._args.window_size, self._args.window_shift)
        terminator = Terminator(args_term)
        terminator.run_terminator(args_term, log)
        log.close()

    def transcript(self):
        """Run Transcript detection"""
        project_creator.create_subfolders(
            self._paths.required_folders("transcript"))
        log = open(os.path.join(self._paths.transcript_output_folder, "log.txt"), "w")
        print("Running transcript detection")
        self.check_multi_files(
                [self._args.annotation_files, self._args.tss_files,
                 self._args.terminator_files],
                ["--annotation_files", "--tss_files", "--terminator_files"], log)
        args_tran = self.args_container.container_transcript(
            self._args.tex_notex, self._args.modify_transcript,
            self._args.length, self._args.annotation_files,
            self._args.height, self._args.width,
            self._args.tolerance, self._args.tolerance_coverage,
            self._args.replicate_tex, self._args.replicate_frag,
            self._paths.transcript_output_folder,
            self._args.tss_files, self._args.tss_tolerance,
            self._args.tex_notex_libs, self._args.frag_libs,
            self._args.compare_feature_genome,
            self._args.terminator_files, self._args.terminator_tolerance,
            self._args.max_length_distribution)
        transcript = TranscriptDetection(args_tran)
        transcript.run_transcript(args_tran, log)

    def utr_detection(self):
        """Run UTR detection."""
        print("Running UTR detection")
        project_creator.create_subfolders(self._paths.required_folders("utr"))
        log = open(os.path.join(self._paths.utr_folder, "log.txt"), "w")
        self.check_multi_files(
            [self._args.annotation_files, self._args.terminator_files,
             self._args.transcript_files, self._args.tss_files],
            ["--annotation_files", "--terminator_files",
             "--transcript_files", "--tss_files"], log)
        args_utr = self.args_container.container_utr(
                self._args.tss_files, self._args.annotation_files,
                self._args.transcript_files, self._args.terminator_files,
                self._args.terminator_tolerance, self._paths.utr_folder,
                self._args.tss_source, self._args.base_5utr,
                self._args.utr_length, self._args.base_3utr)
        utr = UTRDetection(args_utr)
        utr.run_utr_detection(args_utr, log)

    def _check_filter_input(self, files, info, filters, log):
        if files is None:
            print("Error: The {0} has to be provided "
                  "if \"{1}\" in --filter_info!".format(info, filters))
            log.write("The {0} has to be provided "
                      "if \"{1}\" in --filter_info!\n".format(info, filters))
            sys.exit()

    def _check_database(self, database, flag, info, log):
        wrong = False
        if database is None:
            wrong = True
        elif not os.path.isfile(database):
            if (os.path.isfile(database + ".fa")) or (
                    os.path.isfile(database + ".fna")) or (
                    os.path.isfile(database + ".fasta")):
                wrong = False
            else:
                wrong = True
        if wrong:
            print("Error: {0} is required if {1} is in --filter_info. "
                  "But the assignment of {0} is empty or wrong. "
                  "Please check the {0} or remove {1} from "
                  "--filter_info!".format(flag, info))
            log.write("{0} is required if {1} is in --filter_info. "
                      "But the assignment of {0} is empty or wrong. "
                      "Please check the {0} or remove {1} from "
                      "--filter_info!\n".format(flag, info))
            sys.exit()

    def srna_detection(self):
        """sRNA_detection."""
        print("Running sRNA prediction")
        project_creator.create_subfolders(self._paths.required_folders("srna"))
        log = open(os.path.join(self._paths.srna_folder, "log.txt"), "w")
        self.check_multi_files(
                [self._args.annotation_files, self._args.transcript_files,
                 self._args.fasta_files, self._args.sorf_files,
                 self._args.terminator_files, self._args.promoter_tables,
                 self._args.processing_site_files],
                ["--annotation_files", "--transcript_files",
                 "--fasta_files", "--sorf_files", "--terminator_files",
                 "--promoter_tables", "--processing_site_files"], log)
        for info in self._args.filter_info:
            if "sec_str" == info:
                if not self._args.compute_sec_structures:
                    log.write("If you want to use secondary structure to "
                              "filter the false positive, "
                              "--compute_sec_structure need to be switch on.\n")
                    print("Error: --compute_sec_structures is not switch on, "
                          "but sec_str is still in --filter_info.")
                    sys.exit()
                self._check_filter_input(
                        self._args.fasta_files, "fasta file", "sec_str", log)
                for prop in ("rnafold_path", "relplot_path",
                             "mountain_path"):
                    setattr(self._args, prop,
                            self.check_execute_file(getattr(self._args, prop), log))
            elif ("blast_nr" == info) or (
                    "blast_srna"== info):
                for prop in ("blastn_path", "blastx_path", "makeblastdb_path"):
                    setattr(self._args, prop,
                            self.check_execute_file(getattr(self._args, prop), log))
                if ("blast_nr" == info):
                    self._check_database(self._args.nr_database_path,
                                         "--nr_database_path", "blast_nr", log)
                if ("blast_srna" == info):
                    self._check_database(self._args.srna_database_path,
                                         "--srna_database_path", "blast_srna", log)
            elif "sorf" == info:
                self._check_filter_input(
                        self._args.sorf_files, "sORF", "sorf", log)
            elif "term" == info:
                self._check_filter_input(self._args.terminator_files,
                                         "terminator", "term", log)
            elif "promoter" == info:
                self._check_filter_input(self._args.promoter_tables,
                                         "Promoter", "promoter", log)
            elif "tss" == info:
                self._check_filter_input(self._args.tss_files,
                                         "TSS", "tss", log)
            else:
                if "none" != info.lower():
                    print("Error: Please check the --filter_info, "
                          "invalid value was assigned!")
                    log.write("invalid value was assigned to --filter_info.\n")
                    sys.exit()
        log.write("--filter_info and databases are assigned correctly.\n")
        if self._args.utr_derived_srna:
            if self._args.tss_files is None:
                print("Error: The TSS has to be provided "
                      "if you want to compute UTR-derived sRNA!")
                log.write("The TSS has to be provided "
                          "if you want to compute UTR-derived sRNA!\n")
                sys.exit()
        if self._args.search_poly_u != 0:
            if self._args.fasta_files is None:
                print("Error: The fasta files have to be provided "
                      "if you want to extend 3'end of sRNA by "
                      "searching poly U tail!")
                log.write("The fasta files have to be provided "
                          "if you want to extend 3'end of sRNA by "
                          "searching poly U tail!\n")
                sys.exit()
        if ((self._args.nr_format) and (self._args.nr_database_path is None)):
            print("Error: The function for format of nr database was "
                  "switched ont, but no nr database was assigned.")
            sys.exit()
        if ((self._args.srna_format) and (self._args.srna_database_path is None)):
            print("Error: The function for format of srna database was "
                  "switched ont, but no srna database was assigned.")
            sys.exit()
        args_srna = self.args_container.container_srna(
                self._args.rnafold_path, self._args.relplot_path,
                self._args.mountain_path, self._args.blastn_path,
                self._args.blastx_path, self._args.makeblastdb_path,
                self._paths.srna_folder, self._args.utr_derived_srna,
                self._args.annotation_files, self._args.tss_files,
                self._args.transcript_files,
                self._args.tss_intergenic_antisense_tolerance,
                self._args.tss_5utr_tolerance, self._args.tss_3utr_tolerance,
                self._args.tss_intercds_tolerance, self._args.filter_info,
                self._args.processing_site_files, self._args.fasta_files,
                self._args.mountain_plot, self._args.nr_format,
                self._args.srna_format, self._args.srna_database_path,
                self._args.nr_database_path, self._args.cutoff_energy,
                self._args.parallel_blast, self._args.blast_score_srna,
                self._args.blast_score_nr,
                self._args.min_intergenic_tex_coverage,
                self._args.min_intergenic_notex_coverage,
                self._args.min_intergenic_fragmented_coverage,
                self._args.min_complete_5utr_transcript_coverage,
                self._args.min_antisense_tex_coverage,
                self._args.min_antisense_notex_coverage,
                self._args.min_antisense_fragmented_coverage,
                self._args.min_utr_tex_coverage,
                self._args.min_utr_notex_coverage,
                self._args.min_utr_fragmented_coverage,
                self._args.max_length, self._args.min_length,
                self._args.tex_notex_libs, self._args.frag_libs,
                self._args.replicate_tex, self._args.replicate_frag,
                self._args.tex_notex, self._args.blast_e_nr,
                self._args.blast_e_srna, self._args.detect_srna_in_cds,
                self._args.decrease_intergenic_antisense,
                self._args.decrease_utr, self._args.tolerance_intergenic_antisense,
                self._args.tolerance_utr, self._args.cutoff_nr_hit,
                self._args.sorf_files, self._args.overlap_percent_cds,
                self._args.terminator_files,
                self._args.terminator_tolerance_in_srna,
                self._args.terminator_tolerance_out_srna,
                self._args.ignore_hypothetical_protein, self._args.tss_source,
                self._args.min_all_utr_coverage, self._args.promoter_tables,
                self._args.ranking_time_promoter, self._args.promoter_names,
                self._args.compute_sec_structures, self._args.search_poly_u,
                self._args.min_u_poly_u, self._args.mutation_poly_u,
                self._args.exclude_srna_in_annotation_file)
        srna = sRNADetection(args_srna)
        srna.run_srna_detection(args_srna, log)

    def sorf_detection(self):
        """sORF_detection."""
        print("Running sORF prediction")
        project_creator.create_subfolders(
            self._paths.required_folders("sorf"))
        log = open(os.path.join(self._paths.sorf_folder, "log.txt"), "w")
        self.check_multi_files(
                [self._args.transcript_files, self._args.annotation_files,
                 self._args.fasta_files, self._args.srna_files,
                 self._args.tss_files],
                ["--transcript_files", "--annotation_files",
                 "--fasta_files", "--srna_files", "--tss_files"], log)
        args_sorf = self.args_container.container_sorf(
            self._paths.sorf_folder, self._args.utr_derived_sorf,
            self._args.transcript_files,
            self._args.annotation_files,
            self._args.tss_files, self._args.utr_length,
            self._args.min_length, self._args.max_length,
            self._args.cutoff_intergenic_coverage,
            self._args.cutoff_antisense_coverage,
            self._args.cutoff_5utr_coverage,
            self._args.cutoff_3utr_coverage,
            self._args.cutoff_intercds_coverage,
            self._args.fasta_files, self._args.tex_notex_libs,
            self._args.frag_libs, self._args.tex_notex,
            self._args.replicate_tex, self._args.replicate_frag,
            self._args.srna_files,
            self._args.start_codon, self._args.stop_codon,
            self._args.cutoff_base_coverage, self._args.rbs_seq,
            self._args.tolerance_rbs,
            self._args.rbs_not_after_tss, self._args.print_all_combination,
            self._args.best_no_srna, self._args.best_no_tss,
            self._args.ignore_hypothetical_protein,
            self._args.min_rbs_distance, self._args.max_rbs_distance,
            self._args.tolerance_3end, self._args.tolerance_5end,
            self._args.contain_multi_stop)
        sorf = sORFDetection(args_sorf)
        sorf.run_sorf_detection(args_sorf, log)

    def meme(self):
        """promoter detectopn"""
        print("Running promoter detection")
        project_creator.create_subfolders(
            self._paths.required_folders("promoter"))
        log = open(os.path.join(self._paths.promoter_output_folder, "log.txt"), "w")
        self.check_multi_files(
                [self._args.tss_files, self._args.fasta_files],
                ["--tss_files", "--fasta_files"], log)
        if not self._args.tss_source:
            self.check_multi_files([self._args.annotation_files],
                                   ["--annotation_files"], log)
        if (self._args.program == "both") or (
                self._args.program == "meme"):
            self._args.meme_path = self.check_execute_file(self._args.meme_path, log)
        elif (self._args.program == "both") or (
                self._args.program == "glam2"):
            self._args.glam2_path = self.check_execute_file(self._args.glam2_path, log)
        args_pro = self.args_container.container_promoter(
            self._args.meme_path, self._args.glam2_path,
            self._paths.promoter_output_folder, self._args.tex_libs,
            self._args.tss_files, self._args.fasta_files,
            self._args.num_motifs, self._args.nt_before_tss,
            self._args.motif_width, self._args.tss_source,
            self._args.annotation_files, self._args.end_run,
            self._args.combine_all, self._args.e_value,
            self._args.parallels, self._args.program, self._args.use_tss_type)
        meme = MEME(args_pro)
        meme.run_meme(args_pro, log)

    def operon(self):
        """operon detection"""
        print("Running operon detection")
        project_creator.create_subfolders(
            self._paths.required_folders("operon"))
        log = open(os.path.join(self._paths.operon_output_folder, "log.txt"), "w")
        self.check_multi_files(
                [self._args.tss_files, self._args.annotation_files,
                 self._args.transcript_files, self._args.terminator_files],
                ["--tss_files", "--annotation_files",
                 "--transcript_files", "--terminator_files"], log)
        args_op = self.args_container.container_operon(
            self._args.tss_files, self._args.annotation_files,
            self._args.transcript_files, self._args.terminator_files,
            self._args.tss_tolerance, self._args.terminator_tolerance,
            self._args.min_length, self._paths.operon_output_folder,
            self._paths.operon_statistics_folder)
        operon = OperonDetection(args_op)
        operon.run_operon(args_op, log)

    def circrna(self):
        """circRNA detection"""
        print("Running circular RNA prediction")
        project_creator.create_subfolders(
            self._paths.required_folders("circrna"))
        log = open(os.path.join(self._paths.circrna_output_folder,
                                "log.txt"), "w")
        if self._args.read_files:
            self._args.segemehl_path = self.check_execute_file(
                    self._args.segemehl_path, log)
        for prop in ("testrealign_path", "samtools_path"):
            setattr(self._args, prop,
                    self.check_execute_file(getattr(self._args, prop), log))
        self.check_multi_files(
                [self._args.fasta_files, self._args.annotation_files],
                ["--fasta_files", "--annotation_files"], log)
        args_circ = self.args_container.container_circrna(
            self._args.parallels, self._args.fasta_files,
            self._args.annotation_files, self._args.bam_files,
            self._args.read_files, self._paths.circrna_stat_folder,
            self._args.support_reads, self._args.segemehl_path,
            self._args.testrealign_path, self._args.samtools_path,
            self._args.start_ratio, self._args.end_ratio,
            self._args.ignore_hypothetical_protein,
            self._paths.circrna_output_folder)
        circ = CircRNADetection(args_circ)
        circ.run_circrna(args_circ, log)

    def goterm(self):
        """Go term discovery"""
        print("Running GO term mapping")
        project_creator.create_subfolders(
            self._paths.required_folders("go_term"))
        log = open(os.path.join(self._paths.goterm_output_folder, "log.txt"), "w")
        self.check_multi_files(
                [self._args.annotation_files, self._args.transcript_files],
                ["--annotation_files", "--transcript_files"], log)
        self.check_file([self._args.uniprot_id, self._args.go_obo,
                         self._args.goslim_obo],
                        ["--uniprot_id", "--go.obo", "--goslim_obo"], True, log)
        args_go = self.args_container.container_goterm(
            self._args.annotation_files,
            self._paths.goterm_output_folder, self._args.uniprot_id,
            self._args.go_obo, self._args.goslim_obo,
            self._args.transcript_files)
        goterm = GoTermFinding(args_go)
        goterm.run_go_term(args_go, log)

    def srna_target(self):
        """sRNA target prediction"""
        print("Running sRNA target prediction")
        project_creator.create_subfolders(
            self._paths.required_folders("srna_target"))
        log = open(os.path.join(self._paths.starget_output_folder,
                                "log.txt"), "w")
        self.check_multi_files(
                [self._args.fasta_files, self._args.srna_files,
                 self._args.annotation_files],
                ["--fasta_files", "--srna_files",
                 "--annotation_files"], log)
        if "RNAup" in self._args.program:
            self._args.rnaup_path = self.check_execute_file(
                    self._args.rnaup_path, log)
        if "RNAplex" in self._args.program:
            for prop in ("rnaplfold_path", "rnaplex_path"):
                setattr(self._args, prop,
                        self.check_execute_file(getattr(self._args, prop), log))
        if "IntaRNA" in self._args.program:
            self._args.intarna_path = self.check_execute_file(
                    self._args.intarna_path, log)
            if self._args.mode_intarna is None:
                print("Error: --mode_IntaRNA need to be assigned!")
                sys.exit()
        args_tar = self.args_container.container_srna_target(
            self._args.rnaplfold_path, self._args.rnaplex_path,
            self._args.rnaup_path, self._args.intarna_path,
            self._args.annotation_files,
            self._args.fasta_files, self._args.srna_files,
            self._args.query_srnas, self._args.program,
            self._args.interaction_length,
            self._args.window_size_target_rnaplex,
            self._args.span_target_rnaplex,
            self._args.window_size_srna_rnaplfold,
            self._args.span_srna_rnaplfold,
            self._args.unstructured_region_rnaplex_target,
            self._args.unstructured_region_rnaplex_srna,
            self._args.unstructured_region_rnaup,
            self._args.energy_threshold_rnaplex,
            self._args.duplex_distance_rnaplex, self._args.top,
            self._paths.starget_output_folder, self._args.parallels_rnaplex,
            self._args.parallels_rnaup, self._args.parallels_intarna,
            self._args.continue_rnaup,
            self._args.slide_window_size_srna_intarna,
            self._args.max_loop_length_srna_intarna,
            self._args.slide_window_size_target_intarna,
            self._args.max_loop_length_target_intarna,
            self._args.mode_intarna, self._args.potential_target_start,
            self._args.potential_target_end, self._args.target_feature)
        srnatarget = sRNATargetPrediction(args_tar)
        srnatarget.run_srna_target_prediction(args_tar, log)

    def snp(self):
        """SNP transcript detection"""
        print("Running SNP/mutations calling")
        project_creator.create_subfolders(self._paths.required_folders("snp"))
        log = open(os.path.join(self._paths.snp_output_folder,
                                "log.txt"), "w")
        self.check_multi_files(
                [self._args.fasta_files],
                ["--fasta_files"], log)
        if (self._args.bam_type != "related_genome") and (
                self._args.bam_type != "reference_genome"):
            print("Error: Please assign \"related_genome\" or"
                  " \"reference_genome\" to --bam_type!")
            sys.exit()
        if (self._args.ploidy != "haploid") and (
                self._args.ploidy != "diploid"):
            print("Error: Please assign \"haploid\" or"
                  " \"diploid\" to --chromosome_type!")
        if (self._args.caller != "c") and (
                self._args.caller != "m"):
            print("Error: Please assign \"c\" or"
                  " \"m\" to --caller!")
        for prop in ("bcftools_path", "samtools_path"):
            setattr(self._args, prop,
                    self.check_execute_file(getattr(self._args, prop), log))
        args_snp = self.args_container.container_snp(
            self._args.samtools_path, self._args.bcftools_path,
            self._args.bam_type,
            self._args.program, self._args.fasta_files,
            self._args.bam_files,
            self._args.quality, self._args.read_depth_range,
            self._paths.snp_output_folder, self._args.indel_fraction,
            self._args.ploidy, self._args.rg_tag, self._args.caller,
            self._args.filter_tag_info, self._args.dp4_cutoff)
        snp = SNPCalling(args_snp)
        snp.run_snp_calling(args_snp, log)

    def ppi(self):
        """PPI network retrieve"""
        project_creator.create_subfolders(
            self._paths.required_folders("ppi_network"))
        log = open(os.path.join(self._paths.ppi_output_folder,
                                "log.txt"), "w")
        print("Running protein-protein interaction networks prediction")
        self.check_multi_files([self._args.annotation_files],
                               ["--annotation_files"], log)
        self.check_parameter([self._args.query_strains,
                              self._args.species_string],
                             ["--query_strains", "--species_string"], log)
        args_ppi = self.args_container.container_ppi(
            self._args.annotation_files, self._args.query_strains,
            self._args.without_strain_pubmed, self._args.species_string,
            self._args.score, self._paths.ppi_output_folder,
            self._args.node_size, self._args.query)
        ppi = PPINetwork(self._paths.ppi_output_folder)
        ppi.retrieve_ppi_network(args_ppi, log)

    def sublocal(self):
        """Subcellular Localization prediction"""
        print("Running subcellular localization prediction")
        project_creator.create_subfolders(
            self._paths.required_folders("subcellular_localization"))
        log = open(os.path.join(self._paths.sublocal_output_folder,
                                "log.txt"), "w")
        self.check_multi_files(
                [self._args.annotation_files, self._args.fasta_files,
                 self._args.transcript_files],
                ["--annotation_files", "--fasta_files",
                 "--transcript_files"], log)
        if (self._args.bacteria_type != "positive") and (
                self._args.bacteria_type != "negative"):
            print("Error: Please assign \"positive\" or"
                  " \"negative\" to --bacteria_type!")
            sys.exit()
        self._args.psortb_path = self.check_execute_file(self._args.psortb_path, log)
        args_sub = self.args_container.container_sublocal(
            self._args.psortb_path, self._args.annotation_files,
            self._args.fasta_files, self._args.bacteria_type,
            self._args.difference_multi,
            self._paths.sublocal_output_folder, self._args.transcript_files)
        sublocal = SubLocal(args_sub)
        sublocal.run_sub_local(args_sub, log)

    def ribos(self):
        """riboswitch and RNA thermometer prediction"""
        print("Running riboswitch and RNA thermometer prediction")
        log_t = None
        log_r = None
        if (self._args.program == "both"):
            project_creator.create_subfolders(
                    self._paths.required_folders("riboswitch"))
            project_creator.create_subfolders(
                    self._paths.required_folders("thermometer"))
            log_r = open(os.path.join(self._paths.ribos_output_folder,
                                "log.txt"), "w")
            log_t = open(os.path.join(self._paths.thermo_output_folder,
                                "log.txt"), "w")
            self.check_file([self._args.riboswitch_id_file, self._args.rfam_path],
                            ["--riboswitch_id_file", "--rfam_path"], True, log_r)
            self.check_file([self._args.rna_thermometer_id_file,
                             self._args.rfam_path],
                            ["--rna_thermometer_id_file", "--rfam_path"], True, log_t)
            ribos_path = self._paths.ribos_output_folder
            thermo_path = self._paths.thermo_output_folder
        elif (self._args.program == "thermometer"):
            project_creator.create_subfolders(
                    self._paths.required_folders("thermometer"))
            log_t = open(os.path.join(self._paths.thermo_output_folder,
                                "log.txt"), "w")
            self.check_file([self._args.rna_thermometer_id_file,
                             self._args.rfam_path],
                            ["--thermometer_id_file", "--rfam_path"], True, log_t)
            ribos_path = None
            thermo_path = self._paths.thermo_output_folder
        elif (self._args.program == "riboswitch"):
            project_creator.create_subfolders(
                    self._paths.required_folders("riboswitch"))
            log_r = open(os.path.join(self._paths.ribos_output_folder,
                                "log.txt"), "w")
            self.check_file([self._args.riboswitch_id_file, self._args.rfam_path],
                            ["--riboswitch_id_file", "--rfam_path"], True, log_r)
            ribos_path = self._paths.ribos_output_folder
            thermo_path = None
        else:
            log.write("Please assign \"thermometer\", \"riboswitch\" "
                      "or \"both\" in --program.\n")
            print("Error: Please assign \"thermometer\", \"riboswitch\" "
                  "or \"both\" in --program!")
            sys.exit()
        for log in (log_t, log_r):
            if log is not None:
                self.check_multi_files(
                        [self._args.annotation_files, self._args.fasta_files,
                         self._args.tss_files, self._args.transcript_files],
                        ["--annotation_files", "--fasta_files", "--tss_files",
                         "--transcript_files"], log)
                self._args.cmscan_path = self.check_execute_file(self._args.cmscan_path, log)
                self._args.cmpress_path = self.check_execute_file(self._args.cmpress_path, log)
        args_ribo = self.args_container.container_ribos(
            self._args.program, self._args.rna_thermometer_id_file,
            self._args.cmscan_path, self._args.cmpress_path,
            self._args.riboswitch_id_file,
            self._args.annotation_files, self._args.fasta_files,
            self._args.tss_files, self._args.transcript_files,
            self._args.rfam_path, ribos_path,
            thermo_path, self._args.cutoff,
            self._args.output_all, self._paths.database_folder,
            self._args.tolerance, self._args.without_rbs,
            self._args.rbs_seq,
            self._args.tolerance_rbs, self._args.utr_length)
        ribos = Ribos(args_ribo)
        ribos.run_ribos(args_ribo, log_t, log_r)

    def crispr(self):
        """CRISPR prediction"""
        print("Running CRISPR prediction")
        project_creator.create_subfolders(
            self._paths.required_folders("crispr"))
        log = open(os.path.join(self._paths.crispr_output_folder,
                                "log.txt"), "w")
        self.check_multi_files(
                [self._args.fasta_files, self._args.annotation_files],
                ["--fasta_files", "--annotation_files"], log)
        self._args.crt_path = self.check_execute_file(self._args.crt_path, log)
        args_cris = self.args_container.container_cris(
            self._args.fasta_files, self._args.annotation_files,
            self._args.crt_path, self._args.window_size,
            self._args.min_number_repeats, self._args.min_length_repeat,
            self._args.Max_length_repeat, self._args.min_length_spacer,
            self._args.Max_length_spacer, self._paths.crispr_output_folder,
            self._args.ignore_hypothetical_protein)
        cris = Crispr(args_cris)
        cris.run_crispr(args_cris, log)

    def merge(self):
        """Merge all features"""
        print("Merging all features to one gff file")
        merge_folder = os.path.join(self._paths.output_folder,
                                    "merge_all_features")
        self.helper.check_make_folder(merge_folder)
        log = open(os.path.join(merge_folder, "log.txt"), "w")
        other_features = self._args.other_features_files
        self.check_multi_files([[self._args.transcript_file], other_features],
                                ["--transcript_file", "--other_features_files"], log)
        self.check_parameter([self._args.output_prefix], ["--output_prefix"], log)
        run_merge(merge_folder, self._args.transcript_file,
                  self._args.other_features_files,
                  self._args.terminator_tolerance, self._args.tss_tolerance,
                  os.path.join(merge_folder, self._args.output_prefix), log)
        if self._args.source_for_overlapping is not None:
            deal_overlap(merge_folder, self._args.source_for_overlapping)

    def screen(self):
        """generate screenshot"""
        print("Running screenshot generation")
        out_folder = os.path.join(self._args.output_folder, "screenshots")
        if os.path.exists(out_folder):
            print("Error: The {0} already exists!".format(
                  out_folder))
            sys.exit()
        else:
            os.mkdir(out_folder)
        log = open(os.path.join(out_folder, "log.txt"), "w")
        self.check_file([self._args.main_gff, self._args.fasta_file],
                        ["--main_gff", "--fasta_file"], True, log)
        if self._args.side_gffs is not None:
            for gff in (self._args.side_gffs):
                gff = gff.strip()
                if not os.path.isfile(gff):
                    print("Error: The --side_gffs do not exist!")
                    sys.exit()
        if self._args.output_folder is None:
            log.write("No --output_folder can be found.\n")
            print("Error: Please assign --output_folder!")
            sys.exit()
        if (self._args.present != "expand") and (
                self._args.present != "collapse") and (
                self._args.present != "squish"):
            log.write("Please assign \"expand\" or "
                  "\"collapse\" or \"squish\" to --present.\n")
            print("Error: Please assign \"expand\" or "
                  "\"collapse\" or \"squish\" to --present!")
            sys.exit()
        args_sc = self.args_container.container_screen(
            self._args.main_gff, self._args.side_gffs,
            self._args.fasta_file, self._args.height,
            self._args.tex_notex_libs, self._args.frag_libs,
            self._args.present, self._args.output_folder)
        screen = Screen(args_sc, out_folder)
        screen.screenshot(args_sc, log)



import os
import shutil
from annogesiclib.gff3 import Gff3Parser

def get_overlap(anno, source, finals, overlaps, detect, out):
    if (anno.source in source) and (
            anno not in overlaps):
        finals.append(anno)
        detect = True
    return detect

def deal_overlap(out_folder, source):
    gffs = {}
    num = 0
    for gff_file in os.listdir(out_folder):
        if gff_file.endswith(".gff"):
            gff_f = open(os.path.join(out_folder, gff_file), "r")
            for entry in Gff3Parser().entries(gff_f):
                if entry.feature not in gffs.keys():
                    gffs[entry.feature] = []
                gffs[entry.feature].append(entry)
            gff_f.close()
            out = open(os.path.join(out_folder, gff_file + "tmp"), "w")
            finals = []
            overlaps = []
            for feature, annos in gffs.items():
                for anno1 in annos:
                    detect = False
                    for anno2 in annos:
                        if (anno1.seq_id == anno2.seq_id) and (
                            anno1.strand == anno2.strand) and (
                            anno1 != anno2) and (
                            anno1.feature == anno2.feature) and (
                            anno1.source != anno2.source):
                            if ((anno1.start <= anno2.start) and (
                                    anno1.end >= anno2.end)) or (
                                    (anno1.start >= anno2.start) and (
                                    anno1.end <= anno2.end)) or (
                                    (anno1.start <= anno2.start) and (
                                    anno1.end <= anno2.end) and (
                                    anno1.end >= anno2.start)) or (
                                    (anno1.start >= anno2.start) and (
                                    anno1.start <= anno2.end) and (
                                    anno1.end >= anno2.end)):
                                detect = get_overlap(anno1, source, finals,
                                                     overlaps, detect, out)
                                detect = get_overlap(anno2, source, finals,
                                                     overlaps, detect, out)
                                if detect:
                                    overlaps.append(anno1)
                                    overlaps.append(anno2)
                    if (not detect) and (anno1 not in overlaps):
                        finals.append(anno1)
            finals = sorted(finals, key=lambda x: (x.seq_id, x.start,
                                                   x.end, x.strand))
            for final in finals:
                if (final.feature == "region") or (
                        final.feature == "source") or (
                        final.feature == "remark"):
                    out.write(final.info + "\n")
                    break
            for final in finals:
                if (final.feature != "region") and (
                        final.feature != "source"):
                    out.write(final.info + "\n")
            out.close()
            shutil.move(os.path.join(out_folder, gff_file + "tmp"),
                        os.path.join(out_folder, gff_file))



from annogesiclib.gff3 import Gff3Parser


def detect_energy(line, srna):
    duplex = line.split(" ")[0]
    if ("(" in duplex) or (")" in duplex):
        energy = line.split(":")[1].split("(")[1].split("=")[0]
        energy = float(energy.strip())
        if energy < srna["energy"]:
            srna["energy"] = energy


def get_locus(gene):
    if gene is not None:
        if ("locus_tag" in gene.attributes.keys()):
            locus = gene.attributes["locus_tag"]
        else:
            locus = "NA"
    else:
        locus = "NA"
    return locus


def check_parent_gene(cds, genes):
    target_gene = None
    if target_gene is None:
        for gene in genes:
            if (gene.seq_id == cds.seq_id) and (
                    gene.strand == cds.strand):
                if ((cds.start <= gene.start) and (
                        cds.end >= gene.end)) or (
                        (cds.start >= gene.start) and (
                        cds.end <= gene.end)) or (
                        (cds.start <= gene.start) and (
                        cds.end <= gene.end) and (
                        cds.end >= gene.start)) or (
                        (cds.start >= gene.start) and (
                        cds.start <= gene.end) and (
                        cds.end >= gene.end)):
                    target_gene = gene
                if (cds.start == gene.start) and (
                        cds.end == gene.end):
                    target_gene = gene
                    break
    return target_gene


def mod_srna_tar_pos(gff, pos, type_, pre_target, suf_target, length):
    if "NA" not in pos:
        start = int(pos.split(",")[0])
        end = int(pos.split(",")[-1])
        if (gff.strand == "+"):
            if type_ == "srna":
                g_start = gff.start + start - 1
                g_end = gff.start + end - 1
            else:
                if (gff.start - pre_target) <= 0:
                    g_start = start
                    g_end = end
                else:
                    g_start = gff.start - pre_target + start - 1
                    g_end = gff.start - pre_target + end - 1
                if (gff.start - pre_target + end - 1) > length:
                    g_end = length
        else:
            if type_ == "srna":
                g_end = gff.end - start + 1
                g_start = gff.end - end + 1
            else:
                if (gff.end + pre_target) > length:
                    g_start = length - end + 1
                    g_end = length - start + 1
                else:
                    g_start = gff.end + pre_target - end + 1
                    g_end = gff.end + pre_target - start + 1
                if (gff.end + pre_target - end + 1) <= 0:
                    g_start = 1
    else:
        g_start = "NA"
        g_end = "NA"
    return g_start, g_end

def print_rank_one(srnas, out, feature, gffs, srna_gffs, args_tar, length):
    out.write("\t".join(["sRNA", "sRNA_genome", "sRNA_position",
                         "sRNA_interacted_position_" + feature,
                         "sRNA_strand", "Target_genome", "Target_gene_ID",
                         "Target_ID", "Target_locus_tag", "Target_position",
                         "Target_interacted_position_" + feature,
                         "Target_strand", "Energy_" + feature,
                         "Rank_" + feature]) + "\n")
    for method, srna_datas in srnas.items():
        for srna_id, targets in srna_datas.items():
            rank = 0
            sort_targets = sorted(targets, key=lambda k: (k["energy"]))
            for target in sort_targets:
                if (target["tar_pos"] != "NA"):
                    if target["energy"] < 0:
                        rank += 1
                        target["rank"] = rank
                        if (rank <= args_tar.top) and (method == feature):
                            srna_infos = get_srna_name(srna_gffs, srna_id)
                            name = srna_infos[0]
                            srna_info = srna_infos[1]
                            target_info = get_target_info(gffs, target)
                            s_start, s_end = mod_srna_tar_pos(
                                    srna_info, target["srna_pos"], "srna",
                                    args_tar.tar_start, args_tar.tar_end,
                                    length)
                            t_start, t_end = mod_srna_tar_pos(
                                    target_info, target["tar_pos"], "tar",
                                    args_tar.tar_start, args_tar.tar_end,
                                    length)
                            if srna_info is not None:
                                out.write("\t".join([
                                    name, str(srna_info.seq_id),
                                    "-".join([str(srna_info.start),
                                              str(srna_info.end)]),
                                    "-".join([str(s_start), str(s_end)]),
                                    srna_info.strand, target["target_genome"],
                                    target["gene_id"],
                                    target["target_id"],
                                    target["target_locus"],
                                    "-".join([str(target_info.start),
                                              str(target_info.end)]),
                                    "-".join([str(t_start), str(t_end)]),
                                    target_info.strand, str(target["energy"]),
                                    str(target["rank"])]) + "\n")

def extract_pos(line, srnas, method):
    if (line.startswith("(")) or (
            line.startswith(")")) or (
            line.startswith(".")):
        tar_pos = line.split(" : ")[0].strip().split(" ")[-1]
        srna_pos = line.split(" : ")[-1].strip().split(" ")[0]
        srnas["tar_pos"] = tar_pos
        srnas["srna_pos"] = srna_pos

def read_rnaplex(rnaplex, genes, genomes, features, srnas, target_prefixs):
    start = False
    count_seq = 0
    with open(rnaplex, "r") as p_h:
        for line in p_h:
            line = line.strip()
            if line.startswith(">"):
                start = True
                count_seq += 1
                if count_seq == 1:
                    tags = line[1:].split("_")
                    target_locus = "_".join(tags[:-3])
                    target_id = tags[-3]
                    detail = "_".join(tags[-2:])
                    target_genome = target_prefixs[int(tags[0])]
                    gene_id, tar, target_id, target_locus = get_gene_id(
                            detail, genes, genomes, features)
                elif count_seq == 2:
                    srna = line[1:]
                    if srna not in srnas["RNAplex"].keys():
                        srnas["RNAplex"][srna] = []
                    srnas["RNAplex"][srna].append({
                        "target_id": target_id, "gene_id": gene_id,
                        "target_locus": target_locus,
                        "target_genome": target_genome,
                        "detail": detail, "energy": 0})
                    count_seq = 0
            else:
                if start:
                    detect_energy(line, srnas["RNAplex"][srna][-1])
                    extract_pos(line, srnas["RNAplex"][srna][-1], "RNAplex")

def read_rnaup(rnaup, srna_names, srnas, genes, genomes, features, target_prefixs):
    with open(rnaup, "r") as u_h:
        for line in u_h:
            line = line.strip()
            if line.startswith(">"):
                if line[1:] in srna_names:
                    srna = line[1:]
                else:
                    tags = line[1:].split("_")
                    detail = "_".join(tags[-2:])
                    gene_id, tar, target_id, target_locus = get_gene_id(
                            detail, genes, genomes, features)
                    target_genome = target_prefixs[int(tags[0])]
                    if srna in srnas["RNAup"].keys():
                        srnas["RNAup"][srna].append({
                            "target_id": target_id, "target_locus": target_locus,
                            "detail": detail, "energy": 0, "target_genome": target_genome,
                            "gene_id": gene_id})
                    else:
                        srnas["RNAup"][srna] = []
                        srnas["RNAup"][srna].append({
                            "target_id": target_id, "target_locus": target_locus,
                            "detail": detail, "energy": 0, "target_genome": target_genome,
                            "gene_id": gene_id})
            else:
                detect_energy(line, srnas["RNAup"][srna][-1])
                extract_pos(line, srnas["RNAup"][srna][-1], "RNAplex")

def read_intarna(intarna, srnas, genes, genomes, features, target_prefixs):
    with open(intarna, "r") as i_h:
        for line in i_h:
            inter = line.strip().split(";")
            if inter[0] != "id1":
                if len(inter) == 9:
                    srna = inter[3]
                    tags = inter[0].split("_")
                    if (len(tags) >= 5):
                        try:
                            if(int(inter[1]) and int(inter[2]) and 
                                    int(inter[4]) and int(inter[5])):
                                detail = "_".join(tags[-2:])
                                if (len(tags[0])) != 0:
                                    gene_id, tar, target_id, target_locus = get_gene_id(
                                            detail, genes, genomes, features)
                                    target_genome = target_prefixs[int(tags[0])]
                                    if tar is not None:
                                        if (srna not in srnas["IntaRNA"].keys()):
                                            srnas["IntaRNA"][srna] = []
                                        srnas["IntaRNA"][srna].append({
                                            "target_id": target_id,
                                            "target_locus": target_locus,
                                            "detail": detail,
                                            "energy": float(inter[-1]),
                                            "gene_id": gene_id,
                                            "target_genome": target_genome,
                                            "tar_pos": ",".join(inter[1:3]),
                                            "srna_pos": ",".join(inter[4:6])})
                        except:
                            pass

def read_table(gffs, rnaplex, rnaup, intarna, genes, genomes, features, target_prefixs):
    srnas = {"RNAup": {}, "RNAplex": {}, "IntaRNA": {}}
    srna_names = set()
    for gff in gffs:
        if gff.attributes["ID"] not in srna_names:
            srna_names.add(gff.attributes["ID"])
    if rnaplex is not None:
        read_rnaplex(rnaplex, genes, genomes, features, srnas, target_prefixs)
    if rnaup is not None:
        read_rnaup(rnaup, srna_names, srnas, genes, genomes, features, target_prefixs)
    if intarna is not None:
        read_intarna(intarna, srnas, genes, genomes, features, target_prefixs)
    return srnas


def get_gene_id(detail, genes, gffs, features):
    tar = None
    for gff in gffs:
        if gff.feature in features:
            strand = detail.split("_")[-1]
            start = int(detail.split("-")[0])
            end = int(detail.split("-")[1].split("_")[0])
            if (gff.start == start) and (gff.end == end) and (
                    gff.strand == strand):
                tar = gff
                break
    gene_id = "NA"
    target_id = "NA"
    target_locus = "NA"
    if tar is not None:
        if "ID" in tar.attributes.keys():
            target_id = tar.attributes["ID"]
        if "locus_tag" in tar.attributes.keys():
            target_locus = tar.attributes["locus_tag"]
        for gene in genes:
            if "Parent" in tar.attributes.keys():
                if "ID" in gene.attributes.keys():
                    if (gene.attributes["ID"] in
                            tar.attributes["Parent"].split(",")):
                        gene_id = gene.attributes["ID"]
                        return gene_id, tar, target_id, target_locus
            if gene_id == "NA":
                if (gene.seq_id == tar.seq_id) and (
                        gene.strand == tar.strand) and (
                        (tar.start == gene.start) and (
                        tar.end == gene.end)):
                    if "ID" in gene.attributes.keys():
                        gene_id = gene.attributes["ID"]
                        return gene_id, tar, target_id, target_locus
        if gene_id == "NA":
            for gene in genes:
                if (gene.seq_id == tar.seq_id) and (
                        gene.strand == tar.strand):
                    if ((tar.start <= gene.start) and (
                            tar.end >= gene.end)) or (
                            (tar.start >= gene.start) and (
                            tar.end <= gene.end)) or (
                            (tar.start <= gene.start) and (
                            tar.end <= gene.end) and (
                            tar.end >= gene.start)) or (
                            (tar.start >= gene.start) and (
                            tar.start <= gene.end) and (
                            tar.end >= gene.end)):
                        if "ID" in gene.attributes.keys():
                            gene_id = gene.attributes["ID"]
                            return gene_id, tar, target_id, target_locus
    return gene_id, tar, target_id, target_locus


def append_merge_three_methods(name, srna_info, ps_pos, pt_pos, u_data, i_data,
                               srna_m1, target_info, merges):
    merges.append([name, srna_info.seq_id,
                   "-".join([str(srna_info.start), str(srna_info.end)]),
                   ps_pos, u_data["s_pos"], i_data["s_pos"],
                   srna_info.strand, srna_m1["target_genome"], srna_m1["gene_id"],
                   srna_m1["target_id"], srna_m1["target_locus"],
                   "-".join([str(target_info.start), str(target_info.end)]),
                   pt_pos, u_data["t_pos"], i_data["t_pos"],
                   target_info.strand,
                   str(srna_m1["energy"]), str(srna_m1["rank"]),
                   str(u_data["energy"]), str(u_data["rank"]),
                   str(i_data["energy"]), str(i_data["rank"])])


def import_merge(merges, name, srna_info, srna_m1, srna_m2, srna_m3,
                 target_info, pre_target, suf_target, length, num_method):
    ps_start, ps_end = mod_srna_tar_pos(
            srna_info, srna_m1["srna_pos"], "srna", pre_target,
            suf_target, length)
    pt_start, pt_end = mod_srna_tar_pos(
            target_info, srna_m1["tar_pos"], "tar", pre_target,
            suf_target, length)
    ps_pos = "-".join([str(ps_start), str(ps_end)])
    pt_pos = "-".join([str(pt_start), str(pt_end)])
    u_data = {"s_pos": "NA", "t_pos": "NA", "energy": 1000, "rank": "NA"}
    if (srna_m1["detail"] == srna_m2["detail"]):
        us_start, us_end = mod_srna_tar_pos(
                srna_info, srna_m2["srna_pos"], "srna", pre_target,
                suf_target, length)
        ut_start, ut_end = mod_srna_tar_pos(
                target_info, srna_m2["tar_pos"], "tar", pre_target,
                suf_target, length)
        u_data["s_pos"] = "-".join([str(us_start), str(us_end)])
        u_data["t_pos"] = "-".join([str(ut_start), str(ut_end)])
        u_data["energy"] = srna_m2["energy"]
        u_data["rank"] = srna_m2["rank"]
    i_data = {"s_pos": "NA", "t_pos": "NA", "energy": 1000, "rank": "NA"}
    if srna_m3 is not None:
        if (srna_m1["detail"] == srna_m3["detail"]):
            is_start, is_end = mod_srna_tar_pos(
                    srna_info, srna_m3["srna_pos"], "srna", pre_target,
                    suf_target, length)
            it_start, it_end = mod_srna_tar_pos(
                    target_info, srna_m3["tar_pos"], "tar", pre_target,
                    suf_target, length)
            i_data["s_pos"] = "-".join([str(is_start), str(is_end)])
            i_data["t_pos"] = "-".join([str(it_start), str(it_end)])
            i_data["energy"] = srna_m3["energy"]
            i_data["rank"] = srna_m3["rank"]
        append_merge_three_methods(name, srna_info, ps_pos, pt_pos, u_data,
                                   i_data, srna_m1, target_info, merges)
    else:
        if num_method == 2:
            merges.append([name, srna_info.seq_id,
                           "-".join([str(srna_info.start), str(srna_info.end)]),
                           ps_pos, u_data["s_pos"],
                           srna_info.strand, srna_m1["target_genome"], srna_m1["gene_id"],
                           srna_m1["target_id"], srna_m1["target_locus"],
                           "-".join([str(target_info.start), str(target_info.end)]),
                           pt_pos, u_data["t_pos"],
                           target_info.strand,
                           str(srna_m1["energy"]), str(srna_m1["rank"]),
                           str(u_data["energy"]), str(u_data["rank"])])
        elif num_method == 3:
            append_merge_three_methods(name, srna_info, ps_pos, pt_pos, u_data,
                                       i_data, srna_m1, target_info, merges)


def get_srna_name(gffs, srna):
    detect_name = False
    srna_info = None
    for gff in gffs:
        if (gff.attributes["ID"] == srna) and \
           ("Name" in gff.attributes.keys()):
            name = gff.attributes["Name"]
            detect_name = True
            srna_info = gff
            break
        elif (gff.attributes["ID"] == srna):
            srna_info = gff
            break
    if not detect_name:
        name = srna
    return (name, srna_info)


def get_target_info(gffs, target):
    tmp_gff = None
    for gff in gffs:
        if (str(gff.start) + "-" +
            str(gff.end) + "_" + gff.strand) == target["detail"]:
            if gff.feature == "gene":
                target_info = gff
                if ("locus_tag" in gff.attributes.keys()) and (
                        "gene" in gff.attributes.keys()):
                    if (gff.attributes["gene"] not in target["target_locus"]):
                        target["target_locus"] = "|".join([
                            target["target_locus"], gff.attributes["gene"]])
                elif ("locus_tag" in gff.attributes.keys()) and (
                        "Name" in gff.attributes.keys()):
                    if (gff.attributes["Name"] not in target["target_locus"]):
                        target["target_locus"] = "|".join([
                            target["target_locus"], gff.attributes["Name"]])
                return target_info
            else:
                tmp_gff = gff
    if tmp_gff is not None:
        return tmp_gff


def remove_no_rank(merges, index):
    new_merges = []
    for merge in merges:
        if merge[index] != "NA":
            new_merges.append(merge)
    return new_merges


def print_file(merges, out, num_method):
    if num_method == 2:
        merges = remove_no_rank(merges, 15)
        merges = sorted(merges, key=lambda k: (k[0], int(k[15])))
        for merge in merges:
            if float(merge[14]) == 1000:
                merge[14] = "NA"
            if float(merge[16]) == 1000:
                merge[16] = "NA"
            out.write("\t".join(merge) + "\n")
    elif num_method == 3:
        merges = remove_no_rank(merges, 17)
        merges = sorted(merges, key=lambda k: (k[0], int(k[17])))
        for merge in merges:
            if float(merge[16]) == 1000:
                merge[16] = "NA"
            if float(merge[18]) == 1000:
                merge[18] = "NA"
            if float(merge[20]) == 1000:
                merge[20] = "NA"
            out.write("\t".join(merge) + "\n")


def read_gff(filename):
    gffs = []
    genes = []
    for entry in Gff3Parser().entries(open(filename)):
        if entry.feature == "gene":
            genes.append(entry)
        gffs.append(entry)
    gffs = sorted(gffs, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    if len(genes) != 0:
        genes = sorted(genes, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    return gffs, genes


def print_title(out, methods):
    if len(methods) == 2: 
        out.write("\t".join(["sRNA", "sRNA_genome", "sRNA_position",
                             "sRNA_interacted_position_" + methods[0],
                             "sRNA_interacted_position_" + methods[1], "sRNA_strand", "Target_genome",
                             "Target_gene_ID", "Target_ID", "Target_locus_tag", "Target_position",
                             "Target_interacted_position_" + methods[0],
                             "Target_interacted_position_" + methods[1], "Target_strand",
                             "Energy_" + methods[0], "Rank_" + methods[0],
                             "Energy_" + methods[1], "Rank_" + methods[1]]) + "\n")
    if len(methods) == 3:
        out.write("\t".join(["sRNA", "sRNA_genome", "sRNA_position",
                             "sRNA_interacted_position_" + methods[0],
                             "sRNA_interacted_position_" + methods[1],
                             "sRNA_interacted_position_" + methods[2], "sRNA_strand", "Target_genome",
                             "Target_gene_ID", "Target_ID", "Target_locus_tag", "Target_position",
                             "Target_interacted_position_" + methods[0],
                             "Target_interacted_position_" + methods[1],
                             "Target_interacted_position_" + methods[2], "Target_strand",
                             "Energy_" + methods[0], "Rank_" + methods[0],
                             "Energy_" + methods[1], "Rank_" + methods[1],
                             "Energy_" + methods[2], "Rank_" + methods[2]]) + "\n")



def merge_three(srnas, method3, srna, detect, srna_m1, top):
    srna_m3 = None
    if method3 is not None:
        for srna_m3 in srnas[method3][srna]:
            if (srna_m1["detail"] == srna_m3["detail"]) and (
                    "rank" in srna_m3.keys()):
                if (srna_m3["rank"] != "NA"):
                    if (srna_m3["rank"] <= top) and (
                            "print" not in srna_m3.keys()) and (
                            srna_m3["tar_pos"] != "NA"):
                        srna_m3["print"] = True
                        detect["3"] = True
                return srna_m3
    else:
        return srna_m3


def check_method3(srna_m3, srna_m1, method3, srna, srnas):
    if (srna_m3 is not None):
        if "rank" not in srna_m3.keys():
            srna_m3["rank"] = "NA"
            srna_m3["energy"] = 1000
        return srna_m3
    else:
        if method3 is not None:
            for srna_m3 in srnas[method3][srna]:
                if (srna_m1["detail"] == srna_m3["detail"]):
                    return srna_m3


def check_non_overlap(detect, methods, srna_m1, srna_m2, srna_m3, gffs,
                      merges, name, srna_info, args_tar, length,
                      method3, srna, srnas):
    if (not detect["2"] and len(methods) == 2) or (
            ((not detect["2"]) or (not detect["3"]))
            and len(methods) == 3):
        if "rank" not in srna_m2.keys():
            srna_m2["rank"] = "NA"
            srna_m2["energy"] = 1000
        srna_m3 = check_method3(srna_m3, srna_m1, method3, srna, srnas)
        target_info = get_target_info(
            gffs, srna_m1)
        import_merge(
            merges, name, srna_info, srna_m1,
            srna_m2, srna_m3, target_info,
            args_tar.tar_start,
            args_tar.tar_end, length, len(methods))
        srna_m1["print"] = True

def merge_result(srnas, srna_gffs, args_tar, gffs, merges, length, methods):
    '''merge the results based on the ranking of RNAplex'''
    overlaps = []
    method1 = methods[0]
    method2 = methods[1]
    method3 = None
    if len(methods) == 3:
        method3 = methods[2]
    for srna, srna_m1s in srnas[method1].items():
        srna_datas = get_srna_name(srna_gffs, srna)
        name = srna_datas[0]
        srna_info = srna_datas[1]
        for srna_m1 in srna_m1s:
            if ("rank" in srna_m1.keys()):
                if (srna_m1["rank"] != "NA"):
                    if ("print" not in srna_m1.keys()) and (
                            srna_m1["rank"] <= args_tar.top) and (
                            srna_m1["tar_pos"] != "NA"):
                        detect = {"2": False, "3": False}
                        srna_m3 = None
                        for srna_m2 in srnas[method2][srna]:
                            if (srna_m1["detail"] == srna_m2["detail"]):
                                if ("rank" in srna_m2.keys()):
                                    if (srna_m2["rank"] != "NA"):
                                        if (srna_m2["rank"] <= args_tar.top) and (
                                                "print" not in srna_m2.keys()) and (
                                                 srna_m2["tar_pos"] != "NA"):
                                            detect["2"] = True
                                            srna_m3 = merge_three(srnas, method3, srna,
                                                                  detect, srna_m1,
                                                                  args_tar.top)
                                            if (len(methods) == 2) or (
                                                    (len(methods) == 3) and detect["3"]):
                                                target_info = get_target_info(
                                                    gffs, srna_m1)
                                                import_merge(
                                                    overlaps, name, srna_info, srna_m1,
                                                    srna_m2, srna_m3, target_info,
                                                    args_tar.tar_start,
                                                    args_tar.tar_end, length,
                                                    len(methods))
                                                srna_m1["print"] = True
                                                srna_m2["print"] = True
                                                import_merge(
                                                    merges, name, srna_info, srna_m1,
                                                    srna_m2, srna_m3, target_info,
                                                    args_tar.tar_start,
                                                    args_tar.tar_end, length,
                                                    len(methods))
                                break
                        check_non_overlap(detect, methods, srna_m1, srna_m2,
                                          srna_m3, gffs, merges, name, srna_info,
                                          args_tar, length, method3, srna, srnas)
    return overlaps


def compare_rest(srnas, rest, srna_last, srna):
    srna_m3 = None
    for srna_m3 in srnas[rest][srna]:
        if (srna_last["detail"] == srna_m3["detail"]) and (
                "print" not in srna_m3.keys()):
            if ("rank" not in srna_m3.keys()):
                srna_m3["rank"] = "NA"
                srna_m3["energy"] = 1000
            return srna_m3
    return srna_m3


def merge_last(srnas, srna_gffs, args_tar, gffs, merges, length,
               method, ref, num_method, rest, switch):
    '''merge the results based on the ranking of RNAup'''
    for srna, srnas_last in srnas[method].items():
        srna_datas = get_srna_name(srna_gffs, srna)
        name = srna_datas[0]
        srna_info = srna_datas[1]
        for srna_last in srnas_last:
            if "rank" in srna_last.keys():
                if srna_last["rank"] != "NA":
                    if ("print" not in srna_last.keys()) and (
                            srna_last["rank"] <= args_tar.top) and (
                            srna in srnas[ref].keys()) and (
                            srna_last["tar_pos"] != "NA"):
                        for srna_ref in srnas[ref][srna]:
                            if (srna_ref["detail"] == srna_last["detail"]) and (
                                    "print" not in srna_ref.keys()):
                                if ("rank" not in srna_ref.keys()):
                                    srna_ref["rank"] = "NA"
                                    srna_ref["energy"] = 1000
                                else:
                                    target_info = get_target_info(
                                        gffs, srna_last)
                                    if num_method == 2:
                                        import_merge(
                                            merges, name, srna_info, srna_ref,
                                            srna_last, None, target_info,
                                            args_tar.tar_start,
                                            args_tar.tar_end, length, num_method)
                                    elif num_method == 3:
                                        srna_m3 = compare_rest(
                                            srnas, rest, srna_last, srna)
                                        if switch:
                                            import_merge(
                                                merges, name, srna_info,
                                                srna_ref, srna_m3, srna_last,
                                                target_info, args_tar.tar_start,
                                                args_tar.tar_end, length,
                                                num_method)
                                        else:
                                            import_merge(
                                                merges, name, srna_info,
                                                srna_ref, srna_last, srna_m3,
                                                target_info, args_tar.tar_start,
                                                args_tar.tar_end, length,
                                                num_method)
                                        if srna_m3 is not None:
                                            srna_m3["print"] = True
                                    srna_last["print"] = True
                                    srna_ref["print"] = True


def read_fasta(seq_file):
    length = 0
    with open(seq_file) as fh:
        for line in fh:
            line = line.strip()
            if not line.startswith(">"):
                length = length + len(line)
    return length


def merge_srna_target(rnaplex, rnaup, intarna, args_tar, out_rnaplex,
                      out_rnaup, out_intarna, seq_file, output,
                      out_overlap, srna_gff_file, annotation_gff, target_prefixs):
    '''merge the results of RNAup and RNAplex'''
    length = read_fasta(seq_file)
    merges = []
    methods = []
    srna_gffs, NA = read_gff(srna_gff_file)
    gffs, genes = read_gff(annotation_gff)
    srnas = read_table(srna_gffs, rnaplex, rnaup, intarna, genes, gffs,
                       args_tar.features, target_prefixs)
    if out_rnaplex is not None:
        print("Ranking for RNAplex")
        methods.append("RNAplex")
        out_p = open(out_rnaplex, "w")
        print_rank_one(srnas, out_p, "RNAplex", gffs, srna_gffs, args_tar,
                       length)
    if out_rnaup is not None:
        print("Ranking for RNAup")
        methods.append("RNAup")
        out_u = open(out_rnaup, "w")
        print_rank_one(srnas, out_u, "RNAup", gffs, srna_gffs, args_tar,
                       length)
    if out_intarna is not None:
        print("Ranking for IntaRNA")
        methods.append("IntaRNA")
        out_i = open(out_intarna, "w")
        print_rank_one(srnas, out_i, "IntaRNA", gffs, srna_gffs, args_tar,
                       length)
    if (len(args_tar.program) >= 2):
        out_m = open(output, "w")
        out_o = open(out_overlap, "w")
        print_title(out_m, methods)
        print_title(out_o, methods)
        print("Merging now...")
        overlaps = merge_result(srnas, srna_gffs, args_tar, gffs,
                                merges, length, methods)
        if len(methods) == 2:
            merge_last(srnas, srna_gffs, args_tar, gffs, merges, length,
                       methods[1], methods[0], 2, None, False)
        elif len(methods) == 3:
            merge_last(srnas, srna_gffs, args_tar, gffs, merges, length,
                       methods[1], methods[0], 3, methods[2], False)
            merge_last(srnas, srna_gffs, args_tar, gffs, merges, length,
                       methods[2], methods[0], 3, methods[1], True)
        print_file(merges, out_m, len(methods))
        print_file(overlaps, out_o, len(methods))



import os
import shutil
import sys
import random
import csv
from subprocess import Popen, STDOUT
import math
import time
from annogesiclib.gff3 import Gff3Parser
from annogesiclib.converter import Converter
import copy


def compute_stat(stat_value, best, best_para, cores,
                 list_num, out_path, indexs, strain):
    if indexs["change"]:
        indexs["change"] = False
        best = stat_value
        best_para = copy.deepcopy(list_num[-1 * cores + indexs["count"]])
        best_out = open(out_path + "/best_" + strain + ".csv", "w")
        para_line = "_".join(["he", str(best_para["height"]),
                              "rh", str(best_para["re_height"]),
                              "fa", str(best_para["factor"]),
                              "rf", str(best_para["re_factor"]),
                              "bh", str(best_para["base_height"]),
                              "ef", str(best_para["enrichment"]),
                              "pf", str(best_para["processing"])])
        best_out.write("{0}\t{1}\tTP={2}\tTP_rate={3}\tFP={4}\tFP_rate={5}\t"
                       "FN={6}\tMissing_ratio={7}\t".format(
                           (indexs["step"] - cores + 1 + indexs["count"]),
                           para_line, best["tp"], best["tp_rate"], best["fp"],
                           best["fp_rate"], best["fn"], best["missing_ratio"]))
        best_out.close()
    print(", ".join(["Current genome={0}", "Current Parameters:step={1}",
                    "height={2}", "height_reduction={3}", "factor={4}",
                    "factor_reduction={5}", "base_height={6}",
                    "enrichment_factor={7}",
                    "processing_factor={8}"]).format(strain,
                        indexs["step"] - cores + 1 + indexs["count"],
          list_num[-1 * cores + indexs["count"]]["height"],
          list_num[-1 * cores + indexs["count"]]["re_height"],
          list_num[-1 * cores + indexs["count"]]["factor"],
          list_num[-1 * cores + indexs["count"]]["re_factor"],
          list_num[-1 * cores + indexs["count"]]["base_height"],
          list_num[-1 * cores + indexs["count"]]["enrichment"],
          list_num[-1 * cores + indexs["count"]]["processing"]))
    print("Current:TP={0}\tTP_rate={1}\tFP={2}\t"
          "FP_rate={3}\tFN={4}\tMissing_ratio={5}".format(
              stat_value["tp"], stat_value["tp_rate"], stat_value["fp"],
              stat_value["fp_rate"], stat_value["fn"],
              stat_value["missing_ratio"]))
    print("\t".join(["Best Parameters:height={0}", "height_reduction={1}",
                     "factor={2}", "factor_reduction={3}", "base_height={4}",
                     "enrichment_factor={5}", "processing_factor={6}"]).format(
          best_para["height"],
          best_para["re_height"],
          best_para["factor"],
          best_para["re_factor"],
          best_para["base_height"],
          best_para["enrichment"],
          best_para["processing"]))
    print("Best:TP={0}\tTP_rate={1}\tFP={2}\tFP_rate={3}"
          "\tFN={4}\tMissing_ratio={5}".format(
              best["tp"], best["tp_rate"], best["fp"], best["fp_rate"],
              best["fn"], best["missing_ratio"]))
    indexs["count"] += 1
    return (best_para, best)


def scoring_function(best, stat_value, indexs, num_manual):
    '''main scoring function'''
    indexs["change"] = False
    if (stat_value["tp_rate"] == best["tp_rate"]) and (
            stat_value["fp_rate"] == best["fp_rate"]):
        pass
    else:
        if (stat_value["tp_rate"] - best["tp_rate"]) >= 0.1:
            indexs["change"] = True
        elif (best["tp_rate"] - stat_value["tp_rate"]) <= 0.1:
            if (best["tp_rate"] <= stat_value["tp_rate"]) and (
                    best["fp_rate"] >= stat_value["fp_rate"]):
                indexs["change"] = True
            elif num_manual > 100:
                if (stat_value["tp_rate"] - best["tp_rate"] >= 0.01) and (
                        stat_value["fp_rate"] - best["fp_rate"] <= 0.00005):
                    indexs["change"] = True
                elif (best["tp_rate"] - stat_value["tp_rate"] <= 0.01) and (
                        best["fp_rate"] - stat_value["fp_rate"] >= 0.00005):
                    indexs["change"] = True
            tp_diff = float(best["tp"] - stat_value["tp"])
            if tp_diff > 0:
                if float(best["fp"] - stat_value["fp"]) >= 5 * tp_diff:
                    indexs["change"] = True
            elif tp_diff < 0:
                tp_diff = tp_diff * -1
                if float(stat_value["fp"] - best["fp"]) <= 5 * tp_diff:
                    indexs["change"] = True


def check_overlap(overlap, pre_tss, nums, length, manual, predict, pre_pos):
    if overlap:
        if pre_tss:
            pre_tss.attributes["print"] = True
            tss = pre_tss
        else:
            tss = predict
        if (tss.start <= int(length)):
            if (pre_pos != -1):
                if (tss.start - pre_pos != 0):
                    nums["overlap"] += 1
                else:
                    nums["overlap"] += 1
            else:
                nums["overlap"] += 1
        overlap = False
        pre_pos = tss.start
    else:
        if (manual.start <= int(length)):
            nums["manual"] += 1
    return (overlap, pre_pos)


def comparison(manuals, predicts, nums, args_ops, length):
    overlap = False
    pre_pos = -1
    for tss_m in manuals:
        pre_tss = None
        for tss_p in predicts:
            if (tss_p.strand == tss_m.strand) and (
                    (tss_p.seq_id == tss_m.seq_id) or (
                    (tss_p.seq_id == tss_m.seq_id[:-2]) and (
                     tss_m.seq_id[-2] == "."))):
                if (tss_p.start == tss_m.start):
                    tss_p.attributes["print"] = True
                    overlap = True
                    pre_tss = None
                    break
                elif (math.fabs(tss_p.start - tss_m.start) <=
                        args_ops.cluster):
                    overlap = True
                    pre_tss = tss_p
        try:
            datas = check_overlap(overlap, pre_tss, nums, length,
                                  tss_m, tss_p, pre_pos)
            overlap = datas[0]
            pre_pos = datas[1]
        except UnboundLocalError:
            nums = {"overlap": -1, "predict": -1, "manual": -1}
    for tss_p in predicts:
        if tss_p.attributes["print"] is False:
            if (tss_p.start <= int(length)):
                nums["predict"] += 1


def read_predict_manual_gff(gff_file, length):
    num = 0
    gffs = []
    f_h = open(gff_file, "r")
    for entry in Gff3Parser().entries(f_h):
        if (entry.start <= int(length)):
            num += 1
            entry.attributes["print"] = False
            gffs.append(entry)
    f_h.close()
    return num, gffs


def compare_manual_predict(total_step, para_list, gff_files, out_path,
                           out, args_ops, manuals, num_manual, length):
    '''compare manual detected set and prediced set and print to stat.csv'''
#    manual_fh = open(manual, "r")
    stats = []
    count = 0
    total_step = total_step - int(args_ops.cores) + 1
    if num_manual != 0:
        for gff_file in gff_files:
            nums = {"overlap": 0, "predict": 0, "manual": 0}
            para = "_".join(["he", str(para_list[count]["height"]),
                             "rh", str(para_list[count]["re_height"]),
                             "fa", str(para_list[count]["factor"]),
                             "rf", str(para_list[count]["re_factor"]),
                             "bh", str(para_list[count]["base_height"]),
                             "ef", str(para_list[count]["enrichment"]),
                             "pf", str(para_list[count]["processing"])])
            num_predict, predicts = read_predict_manual_gff(gff_file, length)
            comparison(manuals, predicts, nums, args_ops, length)
            out.write("{0}\t{1}\tTP={2}\tTP_rate={3}\t".format(
                      total_step, para, nums["overlap"],
                      float(nums["overlap"]) / float(num_manual)))
            out.write("FP={0}\tFP_rate={1}\tFN={2}"
                      "\tmissing_ratio={3}\n".format(
                          nums["predict"], float(nums["predict"]) / float(
                              int(length) - num_manual),
                          nums["manual"],
                          float(nums["manual"]) / float(num_manual)))
            if nums["overlap"] == -1:
                out.write("No TSS is detected within the range\n")
            stats.append({"tp": nums["overlap"],
                          "tp_rate": float(
                              nums["overlap"]) / float(num_manual),
                          "fp": nums["predict"],
                          "fp_rate": float(nums["predict"]) / float(
                          length - num_manual),
                          "fn": nums["manual"],
                          "missing_ratio": float(
                              nums["manual"]) / float(num_manual)})
            total_step += 1
            count += 1
    return stats


def convert2gff(out_path, gff_files, args_ops, strain):
    for core in range(1, args_ops.cores+1):
        output_folder = os.path.join(
                out_path, "_".join(["MasterTable", str(core)]))
        gff_file = os.path.join(
                output_folder, "_".join(["TSSpredator", str(core) + ".gff"]))
        Converter().convert_mastertable2gff(
                    os.path.join(output_folder, "MasterTable.tsv"),
                    "TSSpredator", args_ops.program,
                    strain, gff_file)
        gff_files.append(gff_file)


def run_TSSpredator(tsspredator_path, config_file):
    folders = config_file.split("/")
    out_path = "/".join(folders[:-1])
    out = open(os.path.join(
        out_path, "TSSpredator_log.txt"), "w")
    p = Popen(["java", "-jar",
               tsspredator_path, config_file], stdout=out, stderr=STDOUT)
    return p


def run_TSSpredator_paralle(config_files, tsspredator_path, processes):
    '''it is for running TSSpredator parallel'''
    for config_file in config_files:
        process = run_TSSpredator(tsspredator_path, config_file)
        processes.append(process)
    for p in processes:
        p.wait()
        if p.stdout:
            p.stdout.close()
        if p.stdin:
            p.stdin.close()
        if p.stderr:
            p.stderr.close()
        try:
            p.kill()
        except OSError:
            pass
    time.sleep(5)


def print_lib(lib_num, lib_list, out, wig_folder, prefix, rep_set):
    for num_id in range(1, lib_num+1):
        cond_list = []
        for lib in lib_list:
            if num_id == lib["condition"]:
                cond_list.append(lib)
        cond_sort_list = sorted(cond_list, key=lambda k: k['replicate'])
        reps = []
        for cond in cond_sort_list:
            out.write("{0}_{1}{2} = {3}/{4}\n".format(
                      prefix, cond["condition"], cond["replicate"],
                      wig_folder, cond["wig"]))
            reps.append(cond["replicate"])
        for rep in sorted(rep_set):
            if rep not in reps:
                out.write("{0}_{1}{2} = \n".format(
                          prefix, cond["condition"], rep))

def assign_dict(lib_datas):
    return {"wig": lib_datas[0],
            "tex": lib_datas[1],
            "condition": int(lib_datas[2]),
            "replicate": lib_datas[3],
            "strand": lib_datas[4]}


def import_lib(wig_folder, rep_set, lib_dict, out, gff,
               list_num_id, fasta, args_ops, strain):
    lib_num = 0
    for lib in args_ops.libs:
        lib_datas = lib.split(":")
        if lib_datas[0].endswith(".wig") is not True:
            print("Error: Exist a not proper wig file!!")
            sys.exit()
        for wig in os.listdir(wig_folder):
            filename = wig.split("_STRAIN_")
            if (filename[0] == lib_datas[0][:-4]) and \
               (filename[1][:-4] == strain):
                lib_datas[0] = wig
            elif (filename[0] == lib_datas[0][:-4]) and \
                 ("." == filename[1][-6]) and \
                 (filename[1][:-6] == strain):
                lib_datas[0] = wig
        if int(lib_datas[2]) > lib_num:
            lib_num = int(lib_datas[2])
        if lib_datas[3] not in rep_set:
            rep_set.add(lib_datas[3])
        if (lib_datas[1] == "tex") and \
           (lib_datas[4] == "+"):
            lib_dict["fp"].append(assign_dict(lib_datas))
        elif (lib_datas[1] == "tex") and \
             (lib_datas[4] == "-"):
            lib_dict["fm"].append(assign_dict(lib_datas))
        elif (lib_datas[1] == "notex") and \
             (lib_datas[4] == "+"):
            lib_dict["np"].append(assign_dict(lib_datas))
        elif (lib_datas[1] == "notex") and \
             (lib_datas[4] == "-"):
            lib_dict["nm"].append(assign_dict(lib_datas))
    for num_id in range(1, lib_num+1):
        os.system("echo '##gff-version 3' > tmp")
        g = open(gff, "r")
        for row in csv.reader(g, delimiter='\t'):
            if not row[0].startswith("#"):
                seq_name = row[0]
                break
        os.system("echo '##sequence-region '" + seq_name + " >> tmp")
        os.system("cat " + gff + ">> tmp")
        g.close()
        shutil.move("tmp", gff)
        out.write("annotation_{0} = {1}\n".format(num_id, gff))
    if args_ops.program.lower() == "tss":
        print_lib(lib_num, lib_dict["fm"], out, wig_folder,
                  "fivePrimeMinus", rep_set)
        print_lib(lib_num, lib_dict["fp"], out, wig_folder,
                  "fivePrimePlus", rep_set)
    elif args_ops.program.lower() == "ps":
        print_lib(lib_num, lib_dict["nm"], out, wig_folder,
                  "fivePrimeMinus", rep_set)
        print_lib(lib_num, lib_dict["np"], out, wig_folder,
                  "fivePrimePlus", rep_set)
    else:
        print("Error: The program name is wrong!!")
        sys.exit()
    for num_id in range(1, lib_num+1):
        out.write("genome_%s = %s\n" % (str(num_id), fasta))
    for num_id in range(1, lib_num+1):
        list_num_id.append(str(num_id))
    return lib_num

def print_repmatch(args_ops, out):
    '''deal with the replicate match'''
    detect_all = False
    for rep in args_ops.replicate:
        if "all" in args_ops.replicate:
            detect_all = True
            match = args_ops.replicate.split("_")[-1]
            out.write("minNumRepMatches = {0}\n".format(match))
            break
    if not detect_all:
        nums = {}
        matchs = {}
        for match in args_ops.replicate:
            lib = match.split("_")[0]
            rep = match.split("_")[-1]
            matchs[lib] = rep
            if rep not in nums.keys():
                nums[rep] = 1
            else:
                nums[rep] += 1
        for rep, num in nums.items():
            if num == max(nums.values()):
                out.write("minNumRepMatches = {0}\n".format(rep))
                max_rep = rep
                break
        for lib, rep in matchs.items():
            if rep != max_rep:
                out.write("minNumRepMatches_{0} = {1}\n".format(
                            lib, rep))


def gen_config(para_list, out_path, core, wig, fasta, gff, args_ops, strain):
    '''generate config file for TSSpredator'''
    files = os.listdir(out_path)
    if "MasterTable_" + str(core) not in files:
        os.mkdir(os.path.join(out_path, "MasterTable_" + str(core)))
    lib_dict = {"fp": [], "fm": [], "nm": [], "np": []}
    rep_set = set()
    list_num_id = []
    filename = os.path.join(out_path, "config_" + str(core) + ".ini")
    out = open(filename, "w")
    out.write("TSSinClusterSelectionMethod = HIGHEST\n")
    out.write("allowedCompareShift = 1\n")
    out.write("allowedRepCompareShift = 1\n")
    lib_num = import_lib(wig, rep_set, lib_dict, out, gff,
                         list_num_id, fasta, args_ops, strain)
    out.write("idList = ")
    out.write(",".join(list_num_id) + "\n")
    out.write("maxASutrLength = 100\n")
    out.write("maxGapLengthInGene = 500\n")
    out.write("maxNormalTo5primeFactor = {0}\n".format(
        para_list["processing"]))
    out.write("maxTSSinClusterDistance = {0}\n".format(args_ops.cluster + 1))
    out.write("maxUTRlength = {0}\n".format(args_ops.utr))
    out.write("min5primeToNormalFactor = {0}\n".format(
        para_list["enrichment"]))
    out.write("minCliffFactor = {0}\n".format(para_list["factor"]))
    out.write("minCliffFactorDiscount = {0}\n".format(para_list["re_factor"]))
    out.write("minCliffHeight = {0}\n".format(para_list["height"]))
    out.write("minCliffHeightDiscount = {0}\n".format(para_list["re_height"]))
    out.write("minNormalHeight = {0}\n".format(para_list["base_height"]))
    print_repmatch(args_ops, out)
    out.write("minPlateauLength = 0\n")
    out.write("mode = cond\n")
    out.write("normPercentile = 0.9\n")
    if (args_ops.program.lower() == "tss"):
        print_lib(lib_num, lib_dict["nm"], out, wig, "normalMinus", rep_set)
        print_lib(lib_num, lib_dict["np"], out, wig, "normalPlus", rep_set)
    elif (args_ops.program.lower() == "ps"):
        print_lib(lib_num, lib_dict["fm"], out, wig, "normalMinus", rep_set)
        print_lib(lib_num, lib_dict["fp"], out, wig, "normalPlus", rep_set)
    out.write("numReplicates = {0}\n".format(len(rep_set)))
    out.write("numberOfDatasets = {0}\n".format(lib_num))
    out.write("outputDirectory = {0}\n".format(
              os.path.join(out_path, "_".join(["MasterTable", str(core)]))))
    for prefix_id in range(len(args_ops.replicate_name)):
        out.write("outputPrefix_{0} = {1}\n".format(
                  prefix_id + 1, args_ops.replicate_name[prefix_id]))
        out.write("outputID_{0} = {1}\n".format(
                  prefix_id + 1, args_ops.output_id))
    out.write("projectName = {0}\n".format(strain))
    out.write("superGraphCompatibility = igb\n")
    out.write("texNormPercentile = 0.5\n")
    out.write("writeGraphs = 0\n")
    out.write("writeNocornacFiles = 0\n")
    out.close()
    return filename


def run_tss_and_stat(indexs, list_num, seeds, diff_h, diff_f,
                     out_path, stat_out, best_para, current_para,
                     wig, fasta, gff, best, num_manual, args_ops, strain,
                     manuals, length, log, set_config, run_tss):
    '''run TSS and do statistics'''
    if indexs["step"] > args_ops.steps + int(args_ops.cores):
        return (True, best_para)
    elif len(list_num) == indexs["length"]:
        indexs["step"] = indexs["step"] - 1
        seeds["pre_seed"].append(seeds["seed"])
    elif (diff_h <= 0) or \
         (diff_f <= 0):
        indexs["step"] = indexs["step"] - 1
        list_num = list_num[:-1]
    else:
        indexs["num"] += 1
        seeds["pre_seed"] = []
        if indexs["num"] == args_ops.cores:
            index = 0
            config_files = []
            gff_files = []
            for para in list_num[-1 * args_ops.cores:]:
                index += 1
                if not set_config:
                    log.write("Checking the process of generating config files.\n")
                config_files.append(gen_config(para, out_path, index, wig,
                                    fasta, gff, args_ops, strain))
                if not set_config:
                    set_config = True
                    log.write("config files can be generated and stored in "
                              "{0} successfully.\n".format(out_path))
            indexs["count"] = 0
            processes = []
            if not run_tss:
                log.write("Checking the setup of TSSpredator.\n")
                log.write("Please make sure your version of TSSpredator "
                          "is at least 1.06.\n")
            run_TSSpredator_paralle(config_files, args_ops.tsspredator_path,
                                    processes)
            if not run_tss:
                run_tss = True
                log.write("TSSpredator is running successfully.\n")
            convert2gff(out_path, gff_files, args_ops, strain)
            stat_values = compare_manual_predict(
                              indexs["step"], list_num[-1 * args_ops.cores:],
                              gff_files, out_path, stat_out, args_ops, manuals,
                              num_manual, length)
            for stat_value in stat_values:
                if indexs["first"]:
                    indexs["first"] = False
                    best = stat_value
                    best_para = copy.deepcopy(list_num[-1 * args_ops.cores +
                                                       indexs["count"]])
                else:
                    scoring_function(best, stat_value, indexs, num_manual)
                datas = compute_stat(stat_value, best, best_para,
                                     args_ops.cores, list_num,
                                     out_path, indexs, strain)
                best_para = datas[0]
                best = datas[1]
            indexs["switch"] += 1
            stat_values = []
            indexs["num"] = 0
    return (False, best_para, best), set_config, run_tss


def minus_process(num_type, new_para, max_num, best_num,
                  actions, list_num, compare):
    '''it is for minus one unit in small change part'''
    if num_type == "base_height":
        new_para[num_type] = new_para[num_type] - 0.001
        new_para[num_type] = float('%.3f' % new_para[num_type])
        while True:
            if new_para[num_type] < 0.0:
                new_para[num_type] = 0.0
                if new_para in list_num:
                    new_para[num_type] = best_num
                    actions["minus"] = True
                    actions["in_or_de"] = 2
                    break
                else:
                    list_num.append(new_para)
                    return new_para[num_type]
            elif (new_para in list_num):
                new_para[num_type] = new_para[num_type] - 0.001
                new_para[num_type] = float('%.3f' % new_para[num_type])
                continue
            else:
                list_num.append(copy.deepcopy(new_para))
                return new_para[num_type]
    else:
        new_para[num_type] = new_para[num_type] - 0.1
        new_para[num_type] = float('%.1f' % new_para[num_type])
        while True:
            if new_para[num_type] <= 0.0:
                new_para[num_type] = best_num
                actions["in_or_de"] = 2
                actions["minus"] = True
                break
            elif (new_para in list_num):
                new_para[num_type] = new_para[num_type] - 0.1
                new_para[num_type] = float('%.1f' % new_para[num_type])
                continue
            elif ((num_type == "factor") or (
                    num_type == "height")) and (
                    new_para[num_type] <= compare):
                new_para[num_type] = best_num
                actions["in_or_de"] = 2
                actions["minus"] = True
                break
            else:
                list_num.append(copy.deepcopy(new_para))
                return new_para[num_type]
    return None


def plus_process(num_type, new_para, max_num,
                 best_num, actions, list_num, compare):
    '''it is for plus one unit in small change part'''
    if num_type == "base_height":
        new_para[num_type] = new_para[num_type] + 0.001
        new_para[num_type] = float('%.3f' % new_para[num_type])
        while True:
            if new_para[num_type] >= max_num:
                new_para[num_type] = best_num
                actions["in_or_de"] = 1
                actions["plus"] = True
                break
            elif (new_para in list_num):
                new_para[num_type] = new_para[num_type] + 0.001
                new_para[num_type] = float('%.3f' % new_para[num_type])
                continue
            else:
                list_num.append(copy.deepcopy(new_para))
                return new_para[num_type]
    else:
        new_para[num_type] = new_para[num_type] + 0.1
        new_para[num_type] = float('%.1f' % new_para[num_type])
        while True:
            if new_para[num_type] >= max_num:
                new_para[num_type] = best_num
                actions["in_or_de"] = 1
                actions["plus"] = True
                break
            elif (new_para in list_num):
                new_para[num_type] = new_para[num_type] + 0.1
                new_para[num_type] = float('%.1f' % new_para[num_type])
                continue
            elif ((num_type == "re_factor") or (
                    num_type == "re_height")) and (
                    new_para[num_type] >= compare):
                new_para[num_type] = best_num
                actions["in_or_de"] = 1
                actions["plus"] = True
                break
            else:
                list_num.append(copy.deepcopy(new_para))
                return new_para[num_type]
    return None


def small_change(max_num, num_type, compare, list_num, best_num, best_para):
    '''add or minus one unit for one parameter in small change part'''
    new_para = copy.deepcopy(best_para)
    actions = {"plus": False, "minus": False}
    step = 0
    if new_para[num_type] >= max_num:
        actions["in_or_de"] = 1
    elif new_para[num_type] <= 0:
        actions["in_or_de"] = 2
    else:
        actions["in_or_de"] = random.randint(0, 9)
    while True:
        step += 1
        if step >= 1000:
            return best_num
        if (actions["plus"] is True) and (actions["minus"] is True):
            new_para[num_type] = best_num
            return new_para[num_type]
        if actions["in_or_de"] % 2 == 0:
            tmp_para = plus_process(num_type, new_para, max_num, best_num,
                                    actions, list_num, compare)
        if actions["in_or_de"] % 2 == 1:
            tmp_para = minus_process(num_type, new_para, max_num, best_num,
                                     actions, list_num, compare)
        if tmp_para is not None:
            return tmp_para


def run_small_change_part(seeds, features, indexs, current_para,
                          best_para, list_num, max_num):
    '''it is for the small change'''
    while True:
        seeds["seed"] = random.randint(0, 6)
        if seeds["seed"] in seeds["pre_seed"]:
            if len(seeds["pre_seed"]) == 7:
                indexs["switch"] += 1
                features["pre_feature"] = features["feature"]
                break
            else:
                continue
        else:
            break
    if seeds["seed"] == 0:
        current_para["height"] = small_change(
                max_num["height"], "height", best_para["re_height"], list_num,
                best_para["height"], best_para)
    elif seeds["seed"] == 1:
        current_para["re_height"] = small_change(
                max_num["re_height"], "re_height", best_para["height"],
                list_num, best_para["re_height"], best_para)
    elif seeds["seed"] == 2:
        current_para["factor"] = small_change(
                max_num["factor"], "factor", best_para["re_factor"], list_num,
                best_para["factor"], best_para)
    elif seeds["seed"] == 3:
        current_para["re_factor"] = small_change(
                max_num["re_factor"], "re_factor", best_para["factor"],
                list_num, best_para["re_factor"], best_para)
    elif seeds["seed"] == 4:
        current_para["base_height"] = small_change(
                max_num["base_height"], "base_height",
                best_para["base_height"], list_num, best_para["base_height"],
                best_para)
    elif seeds["seed"] == 5:
        current_para["enrichment"] = small_change(
                max_num["enrichment"], "enrichment", best_para["enrichment"],
                list_num, best_para["enrichment"], best_para)
    elif seeds["seed"] == 6:
        current_para["processing"] = small_change(
                max_num["processing"], "processing", best_para["processing"],
                list_num, best_para["processing"], best_para)
    return current_para


def gen_large_random(max_num, num_type, compare, list_num, origin_num,
                     best_para, index_large, indexs):
    '''random change two parameters for large change'''
    new_para = copy.deepcopy(best_para)
    step = 0
    while True:
        step += 1
        if step >= 1000000:
            return best_para
        seed = random.randint(0, 6)
        if num_type == index_large[seed]:
            continue
        if num_type == "base_height":
            number = round(random.uniform(0.000, max_num[num_type]), 3)
            number = '%.3f' % number
            number = float(number)
        else:
            number = round(random.uniform(0.1, max_num[num_type]), 1)
            number = '%.1f' % number
            number = float(number)
        if index_large[seed] == "base_height":
            number_par = round(random.uniform(0.000,
                               max_num[index_large[seed]]), 3)
            number_par = '%.3f' % number_par
            number_par = float(number_par)
        else:
            number_par = round(random.uniform(0.1,
                               max_num[index_large[seed]]), 1)
            number_par = '%.1f' % number_par
            number_par = float(number_par)
        new_para[num_type] = number
        new_para[index_large[seed]] = number_par
        if new_para in list_num:
            continue
        else:
            if (new_para["height"] <= new_para["re_height"]) or \
               (new_para["factor"] <= new_para["re_factor"]):
                continue
            else:
                list_num.append(copy.deepcopy(new_para))
                return new_para


def run_large_change_part(seeds, features, indexs, current_para, max_num,
                          best_para, list_num):
    '''it is for the large change'''
    index_large = {0: "height", 1: "re_height", 2: "factor", 3: "re_factor",
                   4: "base_height", 5: "enrichment", 6: "processing"}
    while True:
        seeds["seed"] = random.randint(0, 6)
        if seeds["seed"] in seeds["pre_seed"]:
            if len(seeds["pre_seed"]) == 7:
                features["pre_feature"] = features["feature"]
                indexs["switch"] += 1
                break
            else:
                continue
        else:
            break
    if seeds["seed"] == 0:
        current_para = gen_large_random(
                max_num, "height", best_para["re_height"], list_num,
                best_para["height"], best_para, index_large, indexs)
    elif seeds["seed"] == 1:
        current_para = gen_large_random(
                max_num, "re_height", best_para["height"], list_num,
                best_para["re_height"], best_para, index_large, indexs)
    elif seeds["seed"] == 2:
        current_para = gen_large_random(
                max_num, "factor", best_para["re_factor"], list_num,
                best_para["factor"], best_para, index_large, indexs)
    elif seeds["seed"] == 3:
        current_para = gen_large_random(
                max_num, "re_factor", best_para["factor"], list_num,
                best_para["re_factor"], best_para, index_large, indexs)
    elif seeds["seed"] == 4:
        current_para = gen_large_random(
                max_num, "base_height", best_para["base_height"], list_num,
                best_para["base_height"], best_para, index_large, indexs)
    elif seeds["seed"] == 5:
        current_para = gen_large_random(
                max_num, "enrichment", best_para["enrichment"], list_num,
                best_para["enrichment"], best_para, index_large, indexs)
    elif seeds["seed"] == 6:
        current_para = gen_large_random(
                max_num, "processing", best_para["processing"], list_num,
                best_para["processing"], best_para, index_large, indexs)
    return current_para


def run_random_part(current_para, list_num, max_num, steps, indexs):
    '''it is for the random selection'''
    tmp_random_step = 0
    while True:
        current_para["height"] = round(random.uniform(
                                       0.1, max_num["height"]), 1)
        current_para["re_height"] = round(random.uniform(
                                          0.1, max_num["re_height"]), 1)
        current_para["factor"] = round(random.uniform(
                                       0.1, max_num["factor"]), 1)
        current_para["re_factor"] = round(random.uniform(
                                          0.1, max_num["re_factor"]), 1)
        current_para["enrichment"] = round(random.uniform(
                                           0.1, max_num["enrichment"]), 1)
        current_para["processing"] = round(random.uniform(
                                           0.1, max_num["processing"]), 1)
        current_para["base_height"] = round(random.uniform(
                                            0.000, max_num["base_height"]), 3)
        if (current_para["height"] > current_para["re_height"]) and (
                current_para["factor"] > current_para["re_factor"]) and (
                current_para not in list_num):
            list_num.append(copy.deepcopy(current_para))
            break
        tmp_random_step += 1
        if tmp_random_step >= steps:
            indexs["switch"] += 1
            return None
    return current_para


def optimization_process(indexs, current_para, list_num, max_num, best_para,
                         out_path, stat_out, best, wig, fasta, gff, num_manual,
                         new, args_ops, strain, manuals, length, log):
    '''main part of opimize TSSpredator'''
    features = {"pre_feature": "", "feature": ""}
    seeds = {"pre_seed": [], "seed": 0}
    tmp_step = 0
    log.write("The optimization starts.\n")
    set_config = False
    run_tss = False
    while True:
        if indexs["exist"] is False:
            indexs["exist"] = True
            features["feature"] = ""
        elif (indexs["switch"] % 3 == 0):
            features["feature"] = "r"
            if new:
                start_data(current_para, list_num)
                new = False
            else:
                if features["feature"] != features["pre_feature"]:
                    seeds["pre_seed "] = []
                current_para = run_random_part(current_para, list_num,
                                               max_num, args_ops.steps, indexs)
            if current_para is None:
                tmp_step += 1
        elif (indexs["switch"] % 3 == 1):
            features["feature"] = "l"
            if features["feature"] != features["pre_feature"]:
                seeds["pre_seed"] = []
            current_para = run_large_change_part(
                    seeds, features, indexs, current_para, max_num,
                    best_para, list_num)
        else:
            features["feature"] = "s"
            if features["feature"] != features["pre_feature"]:
                seeds["pre_seed"] = []
            current_para = run_small_change_part(
                    seeds, features, indexs, current_para, best_para,
                    list_num, max_num)
        if current_para is not None:
            diff_h = float(current_para["height"]) - float(
                    current_para["re_height"])
            diff_f = float(current_para["factor"]) - float(
                    current_para["re_factor"])
            datas, set_config, run_tss = run_tss_and_stat(
                    indexs, list_num, seeds, diff_h, diff_f, out_path,
                    stat_out, best_para, current_para, wig, fasta, gff,
                    best, num_manual, args_ops, strain, manuals, length,
                    log, set_config, run_tss)
            tmp_step = 0
        else:
            indexs["step"] = indexs["step"] - 1
        if tmp_step >= 2:
            print("The number of steps may be enough, it "
                  "may not be able to find more parameters\n")
            log.write("The optimization stop because no more combination "
                      "of parameters can be found.\n")
            sys.exit()
        best_para = datas[1]
        if datas[0]:
            break
        else:
            best = datas[2]
        indexs["length"] = len(list_num)
        features["pre_feature"] = features["feature"]
        indexs["step"] += 1
        if indexs["step"] >= args_ops.steps:
            break


def start_data(current_para, list_num):
    '''setup the start parameter as default one'''
    current_para["height"] = 0.3
    current_para["re_height"] = 0.2
    current_para["factor"] = 2.0
    current_para["re_factor"] = 0.5
    current_para["enrichment"] = 2.0
    current_para["processing"] = 1.5
    current_para["base_height"] = 0.000
    list_num.append(copy.deepcopy(current_para))
    return current_para


def extend_data(out_path, best, best_para, step, strain):
    '''extend the data from previous run'''
    print("Current genome is {0}, Extending step from {1}".format(strain, step))
    print("\t".join(["Best Parameters:height={0}", "height_reduction={1}",
                     "factor={2}", "factor_reduction={3}", "base_height={4}",
                     "enrichment_factor={5}", "processing_factor={6}"]).format(
          best_para["height"],
          best_para["re_height"],
          best_para["factor"],
          best_para["re_factor"],
          best_para["base_height"],
          best_para["enrichment"],
          best_para["processing"]))
    print("Best:TP={0}\tTP_rate={1}\tFP={2}\tFP_rate={3}"
          "\tFN={4}\tMissing_ratio={5}".format(
              best["tp"], best["tp_rate"], best["fp"], best["fp_rate"],
              best["fn"], best["missing_ratio"]))
    current_para = copy.deepcopy(best_para)
    return current_para


def load_stat_csv(out_path, list_num, best, best_para, indexs,
                  num_manual, stat_file):
    '''load the statistics from stat.csv'''
    f_h = open(stat_file, "r")
    first_line = True
    line_num = 0
    for row in csv.reader(f_h, delimiter="\t"):
        if (line_num == 0) and (len(row) < 8):
             print("Error: {0} has something wrong, "
                   "please check it or remove it!!!".format(stat_file))
             sys.exit()
        line_num += 1
        paras = row[1].split("_")
        if len(row) == 8:
            prev_stat = {"tp": int(row[2].split("=")[-1]),
                         "tp_rate": float(row[3].split("=")[-1]),
                         "fp": int(row[4].split("=")[-1]),
                         "fp_rate": float(row[5].split("=")[-1])}
            list_num.append({"height": float(paras[1]),
                             "re_height": float(paras[3]),
                             "factor": float(paras[5]),
                             "re_factor": float(paras[7]),
                             "base_height": float(paras[9]),
                             "enrichment": float(paras[11]),
                             "processing": float(paras[13])})
            if first_line:
                first_line = False
                indexs["change"] = True
            else:
                scoring_function(best, prev_stat, indexs, num_manual)
            if indexs["change"]:
                best_para = {"height": float(paras[1]),
                             "re_height": float(paras[3]),
                             "factor": float(paras[5]),
                             "re_factor": float(paras[7]),
                             "base_height": float(paras[9]),
                             "enrichment": float(paras[11]),
                             "processing": float(paras[13])}
                best["tp"] = float(row[2].split("=")[-1])
                best["tp_rate"] = float(row[3].split("=")[-1])
                best["fp"] = float(row[4].split("=")[-1])
                best["fp_rate"] = float(row[5].split("=")[-1])
                best["fn"] = float(row[6].split("=")[-1])
                best["missing_ratio"] = float(row[7].split("=")[-1])
            indexs["step"] = int(row[0]) + 1
    f_h.close()
    return (line_num, best, best_para)


def reload_data(out_path, list_num, best, best_para, indexs,
                num_manual, stat_file, log):
    '''if is based on previous run, it is for reload the previous results'''
    indexs["switch"] = 1
    indexs["exist"] = True
    datas = load_stat_csv(out_path, list_num, best, best_para, indexs,
                          num_manual, stat_file)
    line_num = datas[0]
    best = datas[1]
    best_para = datas[2]
    if len(list_num) > 0:
        indexs["extend"] = True
    else:
        print("Error: {0} has something wrong, "
              "please check it or remove it!!!".format(stat_file))
        log.write(stat_file + " is brocken. Please check it or remove it.\n")
        sys.exit()
    new_line = 0
    new_stat = open("tmp.csv", "w")
    with open(stat_file, "r") as fh:
        for line in fh:
            new_line += 1
            line = line.strip()
            if new_line >= line_num:
                break
            else:
                new_stat.write(line + "\n")
    shutil.move("tmp.csv", stat_file)
    new_stat.close()
    return (best_para, best)


def initiate(args_ops):
    '''setup the dict'''
    max_num = {"height": args_ops.height,
               "re_height": args_ops.height_reduction,
               "factor": args_ops.factor,
               "re_factor": args_ops.factor_reduction,
               "base_height": args_ops.base_height,
               "enrichment": args_ops.enrichment,
               "processing": args_ops.processing}
    best_para = {"height": 0, "re_height": 0,
                 "factor": 0, "re_factor": 0,
                 "base_height": 0, "enrichment": 0,
                 "processing": 0}
    current_para = {"height": 0, "re_height": 0,
                    "factor": 0, "re_factor": 0,
                    "base_height": 0, "enrichment": 0,
                    "processing": 0}
    indexs = {"switch": 0, "extend": False, "exist": False, "step": 0,
              "first": True, "num": 0, "length": 0, "change": False,
              "count": 0}
    return max_num, best_para, current_para, indexs


def check_empty(stat_file):
    empty_file = True
    with open(stat_file) as fh:
        for line in fh:
            if len(line) != 0:
                empty_file = False
                break
    return empty_file

def check_output_id(gff, output_id):
    g = open(gff, "r")
    for row in csv.reader(g, delimiter='\t'):
        if len(row) != 0:
            if (not row[0].startswith("#")):
                tags = row[-1].split(";")
                detect = False
                for tag in tags:
                    if tag.startswith(output_id):
                        detect = True
                if (not detect) and (row[2] == "gene"):
                    print("Warning: --output_id does not exist in "
                          "all genes of annotation gff files.")

def optimization(wig, fasta, gff, args_ops, manual, length, strain, log):
    '''opimize TSSpredator'''
    best = {}
    new = True
    max_num, best_para, current_para, indexs = initiate(args_ops)
    out_path = os.path.join(args_ops.output_folder, "optimized_TSSpredator")
    files = os.listdir(args_ops.output_folder)
    stat_file = os.path.join(out_path, "stat_" + strain + ".csv")
    num_manual, manuals = read_predict_manual_gff(manual, length)
    log.write(manual + " is loaded successfully.\n")
    if len(os.listdir(out_path)) == 1:
        list_num = []
        stat_out = open(stat_file, "w")
    else:
        if (("stat_" + strain + ".csv") in os.listdir(out_path)):
            empty_file = check_empty(stat_file)
            if empty_file:
                os.remove(stat_file)
                list_num = []
                stat_out = open(stat_file, "w")
            else:
                list_num = []
                new = False
                log.write("Checking the previous results.\n")
                datas = reload_data(out_path, list_num, best, best_para, indexs,
                                    num_manual, stat_file, log)
                log.write("The intermediate results of the previous "
                          "optimization is loaded. The optimization will start "
                          "from {0}\n".format(indexs["step"]))
                best_para = datas[0]
                best = datas[1]
                current_para = extend_data(out_path, best,
                                           best_para, indexs["step"], strain)
                stat_out = open(stat_file, "a")
                indexs["first"] = False
        else:
            list_num = []
            stat_out = open(stat_file, "w")
    check_output_id(gff, args_ops.output_id)
    optimization_process(indexs, current_para, list_num, max_num, best_para,
                         out_path, stat_out, best, wig, fasta, gff,
                         num_manual, new, args_ops, strain, manuals, length, log)
    log.write("The optimization is done. The following files are generated:\n")
    for file_ in os.listdir(out_path):
        if not file_.startswith("Master") and not file_.startswith("config"):
            log.write("\t" + file_ + "\n")
    stat_out.close()
